import abc

import numpy as np
from .configs import DefaultConfigs
from .types import DataType
from .constants import DEFAULT_CONSISTENCY_LEVEL
from ..grpc_gen import schema_pb2
from ..exceptions import MilvusException


class LoopBase:
    def __init__(self):
        self.__index = 0

    def __iter__(self):
        return self

    def __getitem__(self, item):
        if isinstance(item, slice):
            _start = item.start or 0
            _end = min(item.stop, self.__len__()) if item.stop else self.__len__()
            _step = item.step or 1

            elements = [self.get__item(i) for i in range(_start, _end, _step)]
            return elements

        if item >= self.__len__():
            raise IndexError("Index out of range")

        return self.get__item(item)

    def __next__(self):
        while self.__index < self.__len__():
            self.__index += 1
            return self.__getitem__(self.__index - 1)

        # iterate stop, raise Exception
        self.__index = 0
        raise StopIteration()

    def __str__(self):
        return str(list(map(str, self.__getitem__(slice(0, 10)))))

    @abc.abstractmethod
    def get__item(self, item):
        raise NotImplementedError()


class LoopCache:
    def __init__(self):
        self._array = []

    def fill(self, index, obj):
        if len(self._array) + 1 < index:
            pass


class FieldSchema:
    def __init__(self, raw):
        self._raw = raw

        #
        self.field_id = 0
        self.name = None
        self.is_primary = False
        self.description = None
        self.auto_id = False
        self.type = DataType.UNKNOWN
        self.indexes = []
        self.params = {}

        ##
        self.__pack(self._raw)

    def __pack(self, raw):
        self.field_id = raw.fieldID
        self.name = raw.name
        self.is_primary = raw.is_primary_key
        self.description = raw.description
        self.auto_id = raw.autoID
        self.type = raw.data_type
        # self.type = DataType(int(raw.type))

        for type_param in raw.type_params:
            if type_param.key == "params":
                import json
                self.params[type_param.key] = json.loads(type_param.value)
            else:
                self.params[type_param.key] = type_param.value
                # maybe we'd better not to check these fields in ORM.
                if type_param.key in ["dim", DefaultConfigs.MaxVarCharLengthKey]:
                    self.params[type_param.key] = int(type_param.value)

        index_dict = {}
        for index_param in raw.index_params:
            if index_param.key == "params":
                import json
                index_dict[index_param.key] = json.loads(index_param.value)
            else:
                index_dict[index_param.key] = index_param.value

        self.indexes.extend([index_dict])

    def dict(self):
        _dict = {
            "field_id": self.field_id,
            "name": self.name,
            "description": self.description,
            "type": self.type,
            "params": self.params or {},
            "is_primary": self.is_primary,
            "auto_id": self.auto_id,
        }
        return _dict


class CollectionSchema:
    def __init__(self, raw):
        self._raw = raw

        #
        self.collection_name = None
        self.description = None
        self.params = {}
        self.fields = []
        self.statistics = {}
        self.auto_id = False  # auto_id is not in collection level any more later
        self.aliases = []
        self.collection_id = 0
        self.consistency_level = DEFAULT_CONSISTENCY_LEVEL  # by default
        self.properties = {}

        #
        if self._raw:
            self.__pack(self._raw)

    def __pack(self, raw):
        self.collection_name = raw.schema.name
        self.description = raw.schema.description
        self.aliases = raw.aliases
        self.collection_id = raw.collectionID

        # keep compatible with older Milvus
        try:
            self.consistency_level = raw.consistency_level
        except Exception:
            self.consistency_level = DEFAULT_CONSISTENCY_LEVEL

        # self.params = dict()
        # TODO: extra_params here
        # for kv in raw.extra_params:
        #     par = ujson.loads(kv.value)
        #     self.params.update(par)
        #     # self.params[kv.key] = kv.value

        self.fields = [FieldSchema(f) for f in raw.schema.fields]

        # for s in raw.statistics:
        #     self.statistics[s.key] = s.value

        self.properties = raw.properties

    def dict(self):
        if not self._raw:
            return {}
        _dict = {
            "collection_name": self.collection_name,
            "auto_id": self.auto_id,
            "description": self.description,
            "fields": [f.dict() for f in self.fields],
            "aliases": self.aliases,
            "collection_id": self.collection_id,
            "consistency_level": self.consistency_level,
            "properties": self.properties,
        }
        return _dict

    def __str__(self):
        return self.dict().__str__()


class Entity:
    def __init__(self, entity_id, entity_row_data, entity_score):
        self._id = entity_id
        self._row_data = entity_row_data
        self._score = entity_score
        self._distance = entity_score

    def __str__(self):
        return f'id: {self._id}, distance: {self._distance}, entity: {self._row_data}'

    def __getattr__(self, item):
        return self.value_of_field(item)

    @property
    def id(self):
        return self._id

    @property
    def fields(self):
        fields = [k for k, v in self._row_data.items()]
        return fields

    def get(self, field):
        return self.value_of_field(field)

    def value_of_field(self, field):
        if field not in self._row_data:
            raise MilvusException(message=f"Field {field} is not in return entity")
        return self._row_data[field]

    def type_of_field(self, field):
        raise NotImplementedError('TODO: support field in Hits')


class Hit:
    def __init__(self, entity_id, entity_row_data, entity_score):
        self._id = entity_id
        self._row_data = entity_row_data
        self._score = entity_score
        self._distance = entity_score

    def __str__(self):
        return str(self.entity)

    __repr__ = __str__

    @property
    def entity(self):
        return Entity(self._id, self._row_data, self._score)

    @property
    def id(self):
        return self._id

    @property
    def distance(self):
        return self._distance

    @property
    def score(self):
        return self._score


class Hits(LoopBase):
    def __init__(self, raw, auto_id, round_decimal=-1):
        super().__init__()
        self._raw = raw
        self._auto_id = auto_id
        if round_decimal != -1:
            self._distances = [round(x, round_decimal) for x in self._raw.scores]
        else:
            self._distances = self._raw.scores

    def __len__(self):
        if self._raw.ids.HasField("int_id"):
            return len(self._raw.ids.int_id.data)
        if self._raw.ids.HasField("str_id"):
            return len(self._raw.ids.str_id.data)
        return 0

    def get__item(self, item):
        if self._raw.ids.HasField("int_id"):
            entity_id = self._raw.ids.int_id.data[item]
        elif self._raw.ids.HasField("str_id"):
            entity_id = self._raw.ids.str_id.data[item]
        else:
            raise MilvusException(message="Unsupported ids type")
        entity_row_data = {}
        if self._raw.fields_data:
            for field_data in self._raw.fields_data:
                if field_data.type == DataType.BOOL:
                    if len(field_data.scalars.bool_data.data) >= item:
                        entity_row_data[field_data.field_name] = field_data.scalars.bool_data.data[item]
                elif field_data.type in (DataType.INT8, DataType.INT16, DataType.INT32):
                    if len(field_data.scalars.int_data.data) >= item:
                        entity_row_data[field_data.field_name] = field_data.scalars.int_data.data[item]
                elif field_data.type == DataType.INT64:
                    if len(field_data.scalars.long_data.data) >= item:
                        entity_row_data[field_data.field_name] = field_data.scalars.long_data.data[item]
                elif field_data.type == DataType.FLOAT:
                    if len(field_data.scalars.float_data.data) >= item:
                        entity_row_data[field_data.field_name] = np.single(field_data.scalars.float_data.data[item])
                elif field_data.type == DataType.DOUBLE:
                    if len(field_data.scalars.double_data.data) >= item:
                        entity_row_data[field_data.field_name] = field_data.scalars.double_data.data[item]
                elif field_data.type == DataType.VARCHAR:
                    if len(field_data.scalars.string_data.data) >= item:
                        entity_row_data[field_data.field_name] = field_data.scalars.string_data.data[item]
                elif field_data.type == DataType.STRING:
                    raise MilvusException(message="Not support string yet")
                    # result[field_data.field_name] = field_data.scalars.string_data.data[index]
                elif field_data.type == DataType.FLOAT_VECTOR:
                    dim = field_data.vectors.dim
                    if len(field_data.vectors.float_vector.data) >= item * dim:
                        start_pos = item * dim
                        end_pos = item * dim + dim
                        entity_row_data[field_data.field_name] = [np.single(x) for x in
                                                                  field_data.vectors.float_vector.data[
                                                                  start_pos:end_pos]]
                elif field_data.type == DataType.BINARY_VECTOR:
                    dim = field_data.vectors.dim
                    if len(field_data.vectors.binary_vector.data) >= item * (dim / 8):
                        start_pos = item * (dim / 8)
                        end_pos = (item + 1) * (dim / 8)
                        entity_row_data[field_data.field_name] = [
                            field_data.vectors.binary_vector.data[start_pos:end_pos]]
        entity_score = self._distances[item]
        return Hit(entity_id, entity_row_data, entity_score)

    @property
    def ids(self):
        if self._raw.ids.HasField("int_id"):
            return self._raw.ids.int_id.data
        if self._raw.ids.HasField("str_id"):
            return self._raw.ids.str_id.data
        return []

    @property
    def distances(self):
        return self._distances


class MutationResult:
    def __init__(self, raw):
        self._raw = raw
        self._primary_keys = []
        self._insert_cnt = 0
        self._delete_cnt = 0
        self._upsert_cnt = 0
        self._timestamp = 0
        self._succ_index = []
        self._err_index = []

        self._pack(raw)

    @property
    def primary_keys(self):
        return self._primary_keys

    @property
    def insert_count(self):
        return self._insert_cnt

    @property
    def delete_count(self):
        return self._delete_cnt

    @property
    def upsert_count(self):
        return self._upsert_cnt

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def succ_count(self):
        return len(self._succ_index)

    @property
    def err_count(self):
        return len(self._err_index)

    @property
    def succ_index(self):
        return self._succ_index

    @property
    def err_index(self):
        return self._err_index

    def __str__(self):
        return f"(insert count: {self._insert_cnt}, delete count: {self._delete_cnt}, upsert count: {self._upsert_cnt}, " \
                f"timestamp: {self._timestamp}, success count: {self.succ_count}, err count: {self.err_count})"

    __repr__ = __str__

    # TODO
    # def error_code(self):
    #     pass
    #
    # def error_reason(self):
    #     pass

    def _pack(self, raw):
        # self._primary_keys = getattr(raw.IDs, raw.IDs.WhichOneof('id_field')).value.data
        which = raw.IDs.WhichOneof("id_field")
        if which == "int_id":
            self._primary_keys = raw.IDs.int_id.data
        elif which == "str_id":
            self._primary_keys = raw.IDs.str_id.data

        self._insert_cnt = raw.insert_cnt
        self._delete_cnt = raw.delete_cnt
        self._upsert_cnt = raw.upsert_cnt
        self._timestamp = raw.timestamp
        self._succ_index = raw.succ_index
        self._err_index = raw.err_index


class QueryResult(LoopBase):
    def __init__(self, raw, auto_id=True):
        super().__init__()
        self._raw = raw
        self._auto_id = auto_id
        self._pack(raw.hits)

    def __len__(self):
        return self._nq

    def _pack(self, raw):
        self._nq = raw.results.num_queries
        self._topk = raw.results.top_k
        self._hits = []
        offset = 0
        for i in range(self._nq):
            hit = schema_pb2.SearchResultData()
            start_pos = offset
            end_pos = offset + raw.results.topks[i]
            hit.scores.append(raw.results.scores[start_pos: end_pos])
            if raw.results.ids.HasField("int_id"):
                hit.ids.append(raw.results.ids.int_id.data[start_pos: end_pos])
            elif raw.results.ids.HasField("str_id"):
                hit.ids.append(raw.results.ids.str_id.data[start_pos: end_pos])
            for field_data in raw.result.fields_data:
                field = schema_pb2.FieldData()
                field.type = field_data.type
                field.field_name = field_data.field_name
                if field_data.type == DataType.BOOL:
                    field.scalars.bool_data.data.extend(field_data.scalars.bool_data.data[start_pos: end_pos])
                elif field_data.type in (DataType.INT8, DataType.INT16, DataType.INT32):
                    field.scalars.int_data.data.extend(field_data.scalars.int_data.data[start_pos: end_pos])
                elif field_data.type == DataType.INT64:
                    field.scalars.long_data.data.extend(field_data.scalars.long_data.data[start_pos: end_pos])
                elif field_data.type == DataType.FLOAT:
                    field.scalars.float_data.data.extend(field_data.scalars.float_data.data[start_pos: end_pos])
                elif field_data.type == DataType.DOUBLE:
                    field.scalars.double_data.data.extend(field_data.scalars.double_data.data[start_pos: end_pos])
                elif field_data.type == DataType.VARCHAR:
                    field.scalars.string_data.data.extend(field_data.scalars.string_data.data[start_pos: end_pos])
                elif field_data.type == DataType.STRING:
                    raise MilvusException(message="Not support string yet")
                    # result[field_data.field_name] = field_data.scalars.string_data.data[index]
                elif field_data.type == DataType.FLOAT_VECTOR:
                    dim = field.vectors.dim
                    field.vectors.dim = dim
                    field.vectors.float_vector.data.extend(
                        field_data.vectors.float_data.data[start_pos * dim: end_pos * dim])
                elif field_data.type == DataType.BINARY_VECTOR:
                    dim = field_data.vectors.dim
                    field.vectors.dim = dim
                    field.vectors.binary_vector.data.extend(field_data.vectors.binary_vector.data[
                                                            start_pos * (dim / 8): end_pos * (dim / 8)])
                hit.fields_data.append(field)
            self._hits.append(hit)
            offset += raw.results.topks[i]

    def get__item(self, item):
        return Hits(self._hits[item], self._auto_id)


class ChunkedQueryResult(LoopBase):
    def __init__(self, raw_list, auto_id=True, round_decimal=-1):
        super().__init__()
        self._raw_list = raw_list
        self._auto_id = auto_id
        self._nq = 0
        self.round_decimal = round_decimal

        self._pack(self._raw_list)

    def __len__(self):
        return self._nq

    def _pack(self, raw_list):
        self._hits = []
        for raw in raw_list:
            nq = raw.results.num_queries
            self._nq += nq
            self._topk = raw.results.top_k
            offset = 0
            for i in range(nq):
                hit = schema_pb2.SearchResultData()
                start_pos = offset
                end_pos = offset + raw.results.topks[i]
                hit.scores.extend(raw.results.scores[start_pos: end_pos])
                if raw.results.ids.HasField("int_id"):
                    hit.ids.int_id.data.extend(raw.results.ids.int_id.data[start_pos: end_pos])
                elif raw.results.ids.HasField("str_id"):
                    hit.ids.str_id.data.extend(raw.results.ids.str_id.data[start_pos: end_pos])
                for field_data in raw.results.fields_data:
                    field = schema_pb2.FieldData()
                    field.type = field_data.type
                    field.field_name = field_data.field_name
                    if field_data.type == DataType.BOOL:
                        field.scalars.bool_data.data.extend(field_data.scalars.bool_data.data[start_pos: end_pos])
                    elif field_data.type in (DataType.INT8, DataType.INT16, DataType.INT32):
                        field.scalars.int_data.data.extend(field_data.scalars.int_data.data[start_pos: end_pos])
                    elif field_data.type == DataType.INT64:
                        field.scalars.long_data.data.extend(field_data.scalars.long_data.data[start_pos: end_pos])
                    elif field_data.type == DataType.FLOAT:
                        field.scalars.float_data.data.extend(field_data.scalars.float_data.data[start_pos: end_pos])
                    elif field_data.type == DataType.DOUBLE:
                        field.scalars.double_data.data.extend(field_data.scalars.double_data.data[start_pos: end_pos])
                    elif field_data.type == DataType.VARCHAR:
                        field.scalars.string_data.data.extend(field_data.scalars.string_data.data[start_pos: end_pos])
                    elif field_data.type == DataType.STRING:
                        raise MilvusException(message="Not support string yet")
                        # result[field_data.field_name] = field_data.scalars.string_data.data[index]
                    elif field_data.type == DataType.FLOAT_VECTOR:
                        dim = field_data.vectors.dim
                        field.vectors.dim = dim
                        field.vectors.float_vector.data.extend(field_data.vectors.float_vector.data[
                                                               start_pos * dim: end_pos * dim])
                    elif field_data.type == DataType.BINARY_VECTOR:
                        dim = field_data.vectors.dim
                        field.vectors.dim = dim
                        field.vectors.binary_vector.data.extend(field_data.vectors.binary_vector.data[
                                                                start_pos * (dim / 8): end_pos * (dim / 8)])
                    hit.fields_data.append(field)
                self._hits.append(hit)
                offset += raw.results.topks[i]

    def get__item(self, item):
        return Hits(self._hits[item], self._auto_id, self.round_decimal)


def _abstract():
    raise NotImplementedError('You need to override this function')


class ConnectIntf:
    """SDK client abstract class

    Connection is a abstract class

    """

    def connect(self, host, port, uri, timeout):
        """
        Connect method should be called before any operations
        Server will be connected after connect return OK
        Should be implemented

        :type  host: str
        :param host: host

        :type  port: str
        :param port: port

        :type  uri: str
        :param uri: (Optional) uri

        :type  timeout: int
        :param timeout:

        :return: Status,  indicate if connect is successful
        """
        _abstract()

    def connected(self):
        """
        connected, connection status
        Should be implemented

        :return: Status,  indicate if connect is successful
        """
        _abstract()

    def disconnect(self):
        """
        Disconnect, server will be disconnected after disconnect return SUCCESS
        Should be implemented

        :return: Status,  indicate if connect is successful
        """
        _abstract()

    def create_table(self, param, timeout):
        """
        Create table
        Should be implemented

        :type  param: TableSchema
        :param param: provide table information to be created

        :type  timeout: int
        :param timeout:

        :return: Status, indicate if connect is successful
        """
        _abstract()

    def has_table(self, table_name, timeout):
        """

        This method is used to test table existence.
        Should be implemented

        :type table_name: str
        :param table_name: table name is going to be tested.

        :type  timeout: int
        :param timeout:

        :return:
            has_table: bool, if given table_name exists

        """
        _abstract()

    def delete_table(self, table_name, timeout):
        """
        Delete table
        Should be implemented

        :type  table_name: str
        :param table_name: table_name of the deleting table

        :type  timeout: int
        :param timeout:

        :return: Status, indicate if connect is successful
        """
        _abstract()

    def add_vectors(self, table_name, records, ids, timeout, **kwargs):
        """
        Add vectors to table
        Should be implemented

        :type  table_name: str
        :param table_name: table name been inserted

        :type  records: list[RowRecord]
        :param records: list of vectors been inserted

        :type  ids: list[int]
        :param ids: list of ids

        :type  timeout: int
        :param timeout:

        :returns:
            Status : indicate if vectors inserted successfully
            ids :list of id, after inserted every vector is given a id
        """
        _abstract()

    def search_vectors(self, table_name, top_k, nprobe, query_records, query_ranges, **kwargs):
        """
        Query vectors in a table
        Should be implemented

        :type  table_name: str
        :param table_name: table name been queried

        :type  query_records: list[RowRecord]
        :param query_records: all vectors going to be queried

        :type  query_ranges: list[Range]
        :param query_ranges: Optional ranges for conditional search.
            If not specified, search whole table

        :type  top_k: int
        :param top_k: how many similar vectors will be searched

        :returns:
            Status:  indicate if query is successful
            query_results: list[TopKQueryResult]
        """
        _abstract()

    def search_vectors_in_files(self, table_name, file_ids, query_records,
                                top_k, nprobe, query_ranges, **kwargs):
        """
        Query vectors in a table, query vector in specified files
        Should be implemented

        :type  table_name: str
        :param table_name: table name been queried

        :type  file_ids: list[str]
        :param file_ids: Specified files id array

        :type  query_records: list[RowRecord]
        :param query_records: all vectors going to be queried

        :type  query_ranges: list[Range]
        :param query_ranges: Optional ranges for conditional search.
            If not specified, search whole table

        :type  top_k: int
        :param top_k: how many similar vectors will be searched

        :returns:
            Status:  indicate if query is successful
            query_results: list[TopKQueryResult]
        """
        _abstract()

    def describe_table(self, table_name, timeout):
        """
        Show table information
        Should be implemented

        :type  table_name: str
        :param table_name: which table to be shown

        :type  timeout: int
        :param timeout:

        :returns:
            Status: indicate if query is successful
            table_schema: TableSchema, given when operation is successful
        """
        _abstract()

    def get_table_row_count(self, table_name, timeout):
        """
        Get table row count
        Should be implemented

        :type  table_name, str
        :param table_name, target table name.

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :returns:
            Status: indicate if operation is successful
            count: int, table row count
        """
        _abstract()

    def show_tables(self, timeout):
        """
        Show all tables in database
        should be implemented

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if this operation is successful
            tables: list[str], list of table names
        """
        _abstract()

    def create_index(self, table_name, index, timeout):
        """
        Create specified index in a table
        should be implemented

        :type  table_name: str
        :param table_name: table name

         :type index: dict
        :param index: index information dict

            example: index = {
                "index_type": IndexType.FLAT,
                "nlist": 18384
            }

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if this operation is successful

        :rtype: Status
        """
        _abstract()

    def server_version(self, timeout):
        """
        Provide server version
        should be implemented

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if operation is successful

            str : Server version

        :rtype: (Status, str)
        """
        _abstract()

    def server_status(self, timeout):
        """
        Provide server status. When cmd !='version', provide 'OK'
        should be implemented

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if operation is successful

            str : Server version

        :rtype: (Status, str)
        """
        _abstract()

    def preload_table(self, table_name, timeout):
        """
        load table to memory cache in advance
        should be implemented

        :param table_name: target table name.
        :type table_name: str

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if operation is successful

        ：:rtype: Status
        """

        _abstract()

    def describe_index(self, table_name, timeout):
        """
        Show index information
        should be implemented

        :param table_name: target table name.
        :type table_name: str

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if operation is successful

            TableSchema: table detail information

        :rtype: (Status, TableSchema)
        """

        _abstract()

    def drop_index(self, table_name, timeout):
        """
        Show index information
        should be implemented

        :param table_name: target table name.
        :type table_name: str

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if operation is successful

        ：:rtype: Status
        """

        _abstract()

    def load_collection(self, collection_name, timeout):
        _abstract()

    def release_collection(self, collection_name, timeout):
        _abstract()

    def load_partitions(self, collection_name, timeout):
        _abstract()

    def release_partitions(self, collection_name, timeout):
        _abstract()



import subprocess
import re
from pkg_resources import get_distribution, DistributionNotFound

__version__ = '0.0.0.dev'

try:
    __version__ = get_distribution('pymilvus').version
except DistributionNotFound:
    # package is not installed
    pass


def get_commit(version="", short=True) -> str:
    """get commit return the commit for a specific version like `xxxxxx.dev12` """

    version_info = r'((\d+)\.(\d+)\.(\d+))((rc)(\d+))?(\.dev(\d+))?'
    # 2.0.0rc9.dev12
    # ('2.0.0', '2', '0', '0', 'rc9', 'rc', '9', '.dev12', '12')
    p = re.compile(version_info)

    target_v = __version__ if version == "" else version
    match = p.match(target_v)

    if match is not None:
        match_version = match.groups()
        if match_version[7] is not None:
            if match_version[4] is not None:
                v = str(int(match_version[6]) - 1)
                target_tag = 'v' + match_version[0] + match_version[5] + v
            else:
                target_tag = 'v' + ".".join(str(int("".join(match_version[1:4])) - 1).split(""))
            target_num = int(match_version[-1])
        elif match_version[4] is not None:
            target_tag = 'v' + match_version[0] + match_version[4]
            target_num = 0
        else:
            target_tag = 'v' + match_version[0]
            target_num = 0
    else:
        return f"Version: {target_v} isn't the right form"

    try:
        cmd = ['git', 'rev-list', '--reverse', '--ancestry-path', f'{target_tag}^..HEAD']
        print(f"git cmd: {' '.join(cmd)}")
        result = subprocess.check_output(cmd).decode('ascii').strip().split('\n')

        length = 7 if short else 40
        return result[target_num][:length]
    except Exception as e:
        return f"Get commit for version {target_v} wrong: {e}"



# Copyright (C) 2019-2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

import abc
from ..client.abstract import Entity


class _IterableWrapper:
    def __init__(self, iterable_obj):
        self._iterable = iterable_obj

    def __iter__(self):
        return self

    def __next__(self):
        return self.on_result(self._iterable.__next__())

    def __getitem__(self, item):
        s = self._iterable.__getitem__(item)
        if isinstance(item, slice):
            _start = item.start or 0
            i_len = self._iterable.__len__()
            _end = min(item.stop, i_len) if item.stop else i_len

            elements = []
            for i in range(_start, _end):
                elements.append(self.on_result(s[i]))
            return elements
        return s

    def __len__(self):
        return self._iterable.__len__()

    @abc.abstractmethod
    def on_result(self, res):
        raise NotImplementedError


# TODO: how to add docstring to method of subclass and don't change the implementation?
#       for example like below:
# class Hits(_IterableWrapper):
#     __init__.__doc__ = """doc of __init__"""
#     __iter__.__doc__ = """doc of __iter__"""
#     __next__.__doc__ = """doc of __next__"""
#     __getitem__.__doc__ = """doc of __getitem__"""
#     __len__.__doc__ = """doc of __len__"""
#
#     def on_result(self, res):
#         return Hit(res)


class DocstringMeta(type):
    def __new__(cls, name, bases, attrs):
        doc_meta = attrs.pop("docstring", None)
        new_cls = super(DocstringMeta, cls).__new__(cls, name, bases, attrs)
        if doc_meta:
            for member_name, member in attrs.items():
                if member_name in doc_meta:
                    member.__doc__ = doc_meta[member_name]
        return new_cls


# for example:
# class Hits(_IterableWrapper, metaclass=DocstringMeta):
#     docstring = {
#         "__init__": """doc of __init__""",
#         "__iter__": """doc of __iter__""",
#         "__next__": """doc of __next__""",
#         "__getitem__": """doc of __getitem__""",
#         "__len__": """doc of __len__""",
#     }
#
#     def on_result(self, res):
#         return Hit(res)


class Hit:
    def __init__(self, hit):
        """
        Construct a Hit object from response. A hit represent a record corresponding to the query.
        """
        self._hit = hit

    @property
    def id(self) -> int:
        """
        Return the id of the hit record.

        :return int:
            The id of the hit record.
        """
        return self._hit.id

    @property
    def entity(self) -> Entity:
        """
        Return the Entity of the hit record.

        :return pymilvus Entity object:
            The entity content of the hit record.
        """
        return self._hit.entity

    @property
    def distance(self) -> float:
        """
        Return the distance between the hit record and the query.

        :return float:
            The distance of the hit record.
        """
        return self._hit.distance

    @property
    def score(self) -> float:
        """
        Return the calculated score of the hit record, now the score is equal to distance.

        :return float:
            The score of the hit record.
        """
        return self._hit.score

    def __str__(self):
        """
        Return the information of hit record.

        :return str:
            The information of hit record.
        """
        return str(self._hit)

    __repr__ = __str__


class Hits:
    def __init__(self, hits):
        """
        Construct a Hits object from response.
        """
        self._hits = hits

    def __iter__(self):
        """
        Iterate the Hits object. Every iteration returns a Hit which represent a record
        corresponding to the query.
        """
        return self

    def __next__(self):
        """
        Iterate the Hits object. Every iteration returns a Hit which represent a record
        corresponding to the query.
        """
        return Hit(self._hits.__next__())

    def __getitem__(self, item):
        """
        Return the kth Hit corresponding to the query.

        :return Hit:
            The kth specified by item Hit corresponding to the query.
        """
        s = self._hits.__getitem__(item)
        if isinstance(item, slice):
            _start = item.start or 0
            i_len = self._hits.__len__()
            _end = min(item.stop, i_len) if item.stop else i_len

            elements = []
            for i in range(_start, _end):
                elements.append(self.on_result(s[i]))
            return elements
        return self.on_result(s)

    def __len__(self) -> int:
        """
        Return the number of hit record.

        :return int:
            The number of hit record.
        """
        return self._hits.__len__()

    def __str__(self):
        return str(list(map(str, self.__getitem__(slice(0, 10)))))

    def on_result(self, res):
        return Hit(res)

    @property
    def ids(self) -> list:
        """
        Return the ids of all hit record.

        :return list[int]:
            The ids of all hit record.
        """
        return self._hits.ids

    @property
    def distances(self) -> list:
        """
        Return the distances of all hit record.

        :return list[float]:
            The distances of all hit record.
        """
        return self._hits.distances


class SearchResult:
    def __init__(self, query_result=None):
        """
        Construct a search result from response.
        """
        self._qs = query_result

    def __iter__(self):
        """
        Iterate the Search Result. Every iteration returns a Hits corresponding to a query.
        """
        return self

    def __next__(self):
        """
        Iterate the Search Result. Every iteration returns a Hits corresponding to a query.
        """
        return self.on_result(self._qs.__next__())

    def __getitem__(self, item):
        """
        Return the Hits corresponding to the nth query.

        :return Hits:
            The hits corresponding to the nth(item) query.
        """
        s = self._qs.__getitem__(item)
        if isinstance(item, slice):
            _start = item.start or 0
            i_len = self._qs.__len__()
            _end = min(item.stop, i_len) if item.stop else i_len

            elements = []
            for i in range(_start, _end):
                elements.append(self.on_result(s[i]))
            return elements
        return self.on_result(s)

    def __len__(self) -> int:
        """
        Return the number of query of Search Result.

        :return int:
            The number of query of search result.
        """
        return self._qs.__len__()

    def __str__(self):
        return str(list(map(str, self.__getitem__(slice(0, 10)))))

    def on_result(self, res):
        return Hits(res)



import time
import datetime
import logging
import functools

import grpc

from .exceptions import MilvusException, MilvusUnavailableException
from .grpc_gen import common_pb2

LOGGER = logging.getLogger(__name__)
WARNING_COLOR = "\033[93m{}\033[0m"


def deprecated(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        dup_msg = "[WARNING] PyMilvus: class Milvus will be deprecated soon, please use Collection/utility instead"
        LOGGER.warning(WARNING_COLOR.format(dup_msg))
        return func(*args, **kwargs)
    return inner


def retry_on_rpc_failure(retry_times=10, initial_back_off=0.01, max_back_off=60, back_off_multiplier=3, retry_on_deadline=True):
    # the default 7 retry_times will cost about 26s
    def wrapper(func):
        @functools.wraps(func)
        @error_handler(func_name=func.__name__)
        @tracing_request()
        def handler(self, *args, **kwargs):
            # This has to make sure every timeout parameter is passing throught kwargs form as `timeout=10`
            _timeout = kwargs.get("timeout", None)
            _retry_on_rate_limit = kwargs.get("retry_on_rate_limit", True)

            retry_timeout = _timeout if _timeout is not None and isinstance(_timeout, int) else None
            counter = 1
            back_off = initial_back_off
            start_time = time.time()

            def timeout(start_time) -> bool:
                """ If timeout is valid, use timeout as the retry limits,
                    If timeout is None, use retry_times as the retry limits.
                """
                if retry_timeout is not None:
                    return time.time() - start_time >= retry_timeout
                return counter > retry_times

            while True:
                try:
                    return func(self, *args, **kwargs)
                except grpc.RpcError as e:
                    # DEADLINE_EXCEEDED means that the task wat not completed
                    # UNAVAILABLE means that the service is not reachable currently
                    # Reference: https://grpc.github.io/grpc/python/grpc.html#grpc-status-code
                    if e.code() != grpc.StatusCode.DEADLINE_EXCEEDED and e.code() != grpc.StatusCode.UNAVAILABLE:
                        raise MilvusException(message=str(e)) from e
                    if not retry_on_deadline and e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                        raise MilvusException(message=str(e)) from e
                    if timeout(start_time):
                        timeout_msg = f"Retry timeout: {retry_timeout}s" if retry_timeout is not None \
                            else f"Retry run out of {retry_times} retry times"

                        if e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                            raise MilvusException(message=f"rpc deadline exceeded: {timeout_msg}") from e
                        if e.code() == grpc.StatusCode.UNAVAILABLE:
                            raise MilvusUnavailableException(message=f"server Unavailable: {timeout_msg}") from e
                        raise MilvusException(message=str(e)) from e

                    if counter > 3:
                        retry_msg = f"[{func.__name__}] retry:{counter}, cost: {back_off:.2f}s, reason: <{e.__class__.__name__}: {e.code()}, {e.details()}>"
                        LOGGER.warning(WARNING_COLOR.format(retry_msg))

                    time.sleep(back_off)
                    back_off = min(back_off * back_off_multiplier, max_back_off)
                except MilvusException as e:
                    if timeout(start_time):
                        timeout_msg = f"Retry timeout: {retry_timeout}s" if retry_timeout is not None \
                            else f"Retry run out of {retry_times} retry times"
                        LOGGER.warning(WARNING_COLOR.format(timeout_msg))
                        raise MilvusException(e.code, f"{timeout_msg}, message={e.message}") from e
                    if _retry_on_rate_limit and e.code == common_pb2.RateLimit:
                        time.sleep(back_off)
                        back_off = min(back_off * back_off_multiplier, max_back_off)
                    else:
                        raise e
                except Exception as e:
                    raise e
                finally:
                    counter += 1

        return handler
    return wrapper


def error_handler(func_name=""):
    def wrapper(func):
        @functools.wraps(func)
        def handler(*args, **kwargs):
            inner_name = func_name
            if inner_name == "":
                inner_name = func.__name__
            record_dict = {}
            try:
                record_dict["RPC start"] = str(datetime.datetime.now())
                return func(*args, **kwargs)
            except MilvusException as e:
                record_dict["RPC error"] = str(datetime.datetime.now())
                LOGGER.error(f"RPC error: [{inner_name}], {e}, <Time:{record_dict}>")
                raise e
            except grpc.FutureTimeoutError as e:
                record_dict["gRPC timeout"] = str(datetime.datetime.now())
                LOGGER.error(f"grpc Timeout: [{inner_name}], <{e.__class__.__name__}: {e.code()}, {e.details()}>, <Time:{record_dict}>")
                raise e
            except grpc.RpcError as e:
                record_dict["gRPC error"] = str(datetime.datetime.now())
                LOGGER.error(f"grpc RpcError: [{inner_name}], <{e.__class__.__name__}: {e.code()}, {e.details()}>, <Time:{record_dict}>")
                raise e
            except Exception as e:
                record_dict["Exception"] = str(datetime.datetime.now())
                LOGGER.error(f"Unexpected error: [{inner_name}], {e}, <Time: {record_dict}>")
                raise MilvusException(message=f"Unexpected error, message=<{str(e)}>") from e
        return handler
    return wrapper


def tracing_request():
    def wrapper(func):
        @functools.wraps(func)
        def handler(self, *args, **kwargs):
            level = kwargs.get("log_level", None)
            req_id = kwargs.get("client_request_id", None)
            if level:
                self.set_onetime_loglevel(level)
            if req_id:
                self.set_onetime_request_id(req_id)
            ret = func(self, *args, **kwargs)
            return ret
        return handler
    return wrapper



import pytest
import grpc
from pymilvus import MilvusException, MilvusUnavailableException
from pymilvus.client.grpc_handler import GrpcHandler
from pymilvus.grpc_gen import milvus_pb2, common_pb2

descriptor = milvus_pb2.DESCRIPTOR.services_by_name['MilvusService']


class TestGrpcHandler:
    @pytest.mark.parametrize("ifHas", [True, False])
    def test_has_collection_no_error(self, channel, client_thread, ifHas):
        handler = GrpcHandler(channel=channel)

        has_collection_future = client_thread.submit(
            handler.has_collection, "fake")

        (invocation_metadata, request, rpc) = (
            channel.take_unary_unary(descriptor.methods_by_name['DescribeCollection']))
        rpc.send_initial_metadata(())

        reason = "" if ifHas else "can't find collection"
        error_code = common_pb2.Success if ifHas else common_pb2.UnexpectedError

        expected_result = milvus_pb2.DescribeCollectionResponse(
            status=common_pb2.Status(
                error_code=error_code,
                reason=reason),
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, '')

        got_result = has_collection_future.result()
        assert got_result is ifHas

    def test_has_collection_error(self, channel, client_thread):
        handler = GrpcHandler(channel=channel)

        has_collection_future = client_thread.submit(
            handler.has_collection, "fake")

        (invocation_metadata, request, rpc) = (
            channel.take_unary_unary(descriptor.methods_by_name['DescribeCollection']))
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.DescribeCollectionResponse(
            status=common_pb2.Status(
                error_code=common_pb2.UnexpectedError,
                reason="other reason"),
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, '')

        with pytest.raises(MilvusException):
            has_collection_future.result()

    def test_has_collection_Unavailable_exception(self, channel, client_thread):
        handler = GrpcHandler(channel=channel)
        channel.close()

        # Retry is unable to test
        has_collection_future = client_thread.submit(
            handler.has_collection, "fake", timeout=0)

        (invocation_metadata, request, rpc) = (
            channel.take_unary_unary(descriptor.methods_by_name['DescribeCollection']))
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.DescribeCollectionResponse()

        rpc.terminate(expected_result, (), grpc.StatusCode.UNAVAILABLE, 'server Unavailable')

        with pytest.raises(MilvusUnavailableException):
            has_collection_future.result()

    def test_get_server_version_error(self, channel, client_thread):
        handler = GrpcHandler(channel=channel)

        get_version_future = client_thread.submit(
            handler.get_server_version)

        (invocation_metadata, request, rpc) = (
            channel.take_unary_unary(descriptor.methods_by_name['GetVersion']))
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.GetVersionResponse(
            status=common_pb2.Status(
                error_code=common_pb2.UnexpectedError,
                reason="unexpected error"),
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, '')

        with pytest.raises(MilvusException):
            get_version_future.result()

    def test_get_server_version_error(self, channel, client_thread):
        version = "2.2.0"
        handler = GrpcHandler(channel=channel)

        get_version_future = client_thread.submit(
            handler.get_server_version)

        (invocation_metadata, request, rpc) = (
            channel.take_unary_unary(descriptor.methods_by_name['GetVersion']))
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.GetVersionResponse(
            status=common_pb2.Status(error_code=common_pb2.Success),
            version=version,
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, '')

        got_result = get_version_future.result()
        assert got_result == version

    @pytest.mark.parametrize("_async", [True])
    def test_flush_all(self, channel, client_thread, _async):
        handler = GrpcHandler(channel=channel)

        flush_all_future = client_thread.submit(
            handler.flush_all, _async=_async, timeout=10)

        (invocation_metadata, request, rpc) = (
            channel.take_unary_unary(descriptor.methods_by_name['FlushAll']))
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.FlushAllResponse(
            status=common_pb2.Status(error_code=common_pb2.Success),
            flush_all_ts=100,
        )

        rpc.terminate(expected_result, (), grpc.StatusCode.OK, '')
        assert flush_all_future is not None

    def test_get_flush_all_state(self, channel, client_thread):
        handler = GrpcHandler(channel=channel)

        flushed = client_thread.submit(
            handler.get_flush_all_state, flush_all_ts=100)

        (invocation_metadata, request, rpc) = (
            channel.take_unary_unary(descriptor.methods_by_name['GetFlushAllState']))
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.GetFlushStateResponse(
            status=common_pb2.Status(error_code=common_pb2.Success),
            flushed=True,
        )

        rpc.terminate(expected_result, (), grpc.StatusCode.OK, '')
        assert flushed.result() is True



# Copyright (C) 2019-2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

import copy
from typing import List
import pandas
from pandas.api.types import is_list_like

from .constants import COMMON_TYPE_PARAMS
from .types import DataType, map_numpy_dtype_to_datatype, infer_dtype_bydata
from ..exceptions import (
    CannotInferSchemaException,
    DataTypeNotSupportException,
    PrimaryKeyException,
    FieldsTypeException,
    FieldTypeException,
    AutoIDException,
    ExceptionsMessage,
    DataNotMatchException,
    SchemaNotReadyException,
)


class CollectionSchema:
    def __init__(self, fields, description="", **kwargs):
        if not isinstance(fields, list):
            raise FieldsTypeException(message=ExceptionsMessage.FieldsType)
        self._fields = [copy.deepcopy(field) for field in fields]
        primary_field = kwargs.get("primary_field", None)
        for field in self._fields:
            if not isinstance(field, FieldSchema):
                raise FieldTypeException(message=ExceptionsMessage.FieldType)
            if primary_field == field.name:
                field.is_primary = True
        self._primary_field = None
        for field in self._fields:
            if field.is_primary:
                if primary_field is not None and primary_field != field.name:
                    raise PrimaryKeyException(message=ExceptionsMessage.PrimaryKeyOnlyOne)
                self._primary_field = field
                primary_field = field.name

        if self._primary_field is None:
            raise PrimaryKeyException(message=ExceptionsMessage.PrimaryKeyNotExist)

        if self._primary_field.dtype not in [DataType.INT64, DataType.VARCHAR]:
            raise PrimaryKeyException(message=ExceptionsMessage.PrimaryKeyType)

        self._auto_id = kwargs.get("auto_id", None)
        if "auto_id" in kwargs:
            if not isinstance(self._auto_id, bool):
                raise AutoIDException(message=ExceptionsMessage.AutoIDType)
            if self._primary_field.auto_id is not None and self._primary_field.auto_id != self._auto_id:
                raise AutoIDException(message=ExceptionsMessage.AutoIDInconsistent)
            self._primary_field.auto_id = self._auto_id
            if self._primary_field.auto_id and self._primary_field.dtype == DataType.VARCHAR:
                raise AutoIDException(message=ExceptionsMessage.AutoIDFieldType)
        else:
            if self._primary_field.auto_id is None:
                self._primary_field.auto_id = False
            self._auto_id = self._primary_field.auto_id

        self._description = description
        self._kwargs = copy.deepcopy(kwargs)

    def __repr__(self):
        _dict = {
            "auto_id": self.auto_id,
            "description": self._description,
            "fields": self._fields,
        }
        r = ["{\n"]
        s = "  {}: {}\n"
        for k, v in _dict.items():
            r.append(s.format(k, v))
        r.append("}")
        return "".join(r)

    def __len__(self):
        return len(self.fields)

    def __eq__(self, other):
        """
        The order of the fields of schema must be consistent.
        """
        return self.to_dict() == other.to_dict()

    @classmethod
    def construct_from_dict(cls, raw):
        fields = [FieldSchema.construct_from_dict(field_raw) for field_raw in raw['fields']]
        return CollectionSchema(fields, raw.get('description', ""))

    @property
    # TODO:
    def primary_field(self):
        return self._primary_field

    @property
    def fields(self):
        """
        Returns the fields about the CollectionSchema.

        :return list:
            List of FieldSchema, return when operation is successful.

        :example:
            >>> from pymilvus import FieldSchema, CollectionSchema, DataType
            >>> field = FieldSchema("int64", DataType.INT64, description="int64", is_primary=True)
            >>> schema = CollectionSchema(fields=[field])
            >>> schema.fields
            [<pymilvus.schema.FieldSchema object at 0x7fd3716ffc50>]
        """
        return self._fields

    @property
    def description(self):
        """
        Returns a text description of the CollectionSchema.

        :return str:
            CollectionSchema description text, return when operation is successful.

        :example:
            >>> from pymilvus import FieldSchema, CollectionSchema, DataType
            >>> field = FieldSchema("int64", DataType.INT64, description="int64", is_primary=True)
            >>> schema = CollectionSchema(fields=[field], description="test get description")
            >>> schema.description
            'test get description'
        """
        return self._description

    @property
    def auto_id(self):
        """
        Whether the primary keys are automatically generated.

        :return bool:
            * True: If the primary keys are automatically generated,
            * False: Otherwise.

        :example:
            >>> from pymilvus import FieldSchema, CollectionSchema, DataType
            >>> field = FieldSchema("int64", DataType.INT64, description="int64", is_primary=True)
            >>> schema = CollectionSchema(fields=[field])
            >>> schema.auto_id
            False
        """
        return self._auto_id

    def to_dict(self):
        _dict = {
            "auto_id": self.auto_id,
            "description": self._description,
            "fields": [s.to_dict() for s in self._fields],
        }
        return _dict


class FieldSchema:
    def __init__(self, name: str, dtype: DataType, description="", **kwargs):
        self.name = name
        try:
            dtype = DataType(dtype)
        except ValueError:
            raise DataTypeNotSupportException(message=ExceptionsMessage.FieldDtype) from None
        if dtype == DataType.UNKNOWN:
            raise DataTypeNotSupportException(message=ExceptionsMessage.FieldDtype)
        self._dtype = dtype
        self._description = description
        self._type_params = {}
        self._kwargs = copy.deepcopy(kwargs)
        if not isinstance(kwargs.get("is_primary", False), bool):
            raise PrimaryKeyException(message=ExceptionsMessage.IsPrimaryType)
        self.is_primary = kwargs.get("is_primary", False)
        self.auto_id = kwargs.get("auto_id", None)
        if "auto_id" in kwargs:
            if not isinstance(self.auto_id, bool):
                raise AutoIDException(message=ExceptionsMessage.AutoIDType)
            if not self.is_primary and self.auto_id:
                raise PrimaryKeyException(message=ExceptionsMessage.AutoIDOnlyOnPK)

        self._parse_type_params()

    def __repr__(self):
        r = ["{\n"]
        s = "    {}: {}\n"
        for k, v in self.to_dict().items():
            r.append(s.format(k, v))
        r.append("  }")
        return "".join(r)

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        return self.construct_from_dict(self.to_dict())

    def _parse_type_params(self):
        # update self._type_params according to self._kwargs
        if self._dtype not in (DataType.BINARY_VECTOR, DataType.FLOAT_VECTOR, DataType.VARCHAR,):
            return
        if not self._kwargs:
            return
        # currently only support ndim
        if self._kwargs:
            for k in COMMON_TYPE_PARAMS:
                if k in self._kwargs:
                    if self._type_params is None:
                        self._type_params = {}
                    self._type_params[k] = self._kwargs[k]

    @classmethod
    def construct_from_dict(cls, raw):
        kwargs = {}
        kwargs.update(raw.get("params", {}))
        kwargs['is_primary'] = raw.get("is_primary", False)
        if raw.get("auto_id", None) is not None:
            kwargs['auto_id'] = raw.get("auto_id", None)
        return FieldSchema(raw['name'], raw['type'], raw.get("description", ""), **kwargs)

    def to_dict(self):
        _dict = {
            "name": self.name,
            "description": self._description,
            "type": self.dtype,
        }
        if self._type_params:
            _dict["params"] = copy.deepcopy(self.params)
        if self.is_primary:
            _dict["is_primary"] = True
            _dict["auto_id"] = self.auto_id
        return _dict

    def __getattr__(self, item):
        if self._type_params and item in self._type_params:
            return self._type_params[item]
        return None

    def __eq__(self, other):
        if not isinstance(other, FieldSchema):
            return False
        return self.to_dict() == other.to_dict()

    @property
    def description(self):
        """
        Returns the text description of the FieldSchema.

        :return str:
            FieldSchema description text, returned when the operation is successful.

        :example:
        >>> from pymilvus import FieldSchema, DataType
        >>> field = FieldSchema("int64", DataType.INT64, description="int64", is_primary=False)
        >>> field.description
        'int64'
        """
        return self._description

    @property
    def params(self):
        """
        Returns the parameters of the field.

        :return list:
            List of the parameter.

        :example:
        >>> from pymilvus import FieldSchema, DataType
        >>> field = FieldSchema("int64", DataType.INT64, description="int64", is_primary=False)
        >>> field.params
        {}
        >>> fvec_field = FieldSchema("fvec", DataType.FLOAT_VECTOR, is_primary=False, dim=128)
        >>> fvec_field.params
        {'dim': 128}
        """
        return self._type_params

    @property
    def dtype(self) -> DataType:
        return self._dtype


def check_insert_data_schema(schema: CollectionSchema, data: [List[List], pandas.DataFrame]) -> None:
    """ check if the insert data is consist with the collection schema

    Args:
        schema (CollectionSchema): the schema of the collection
        data (List[List], pandas.DataFrame): the data to be inserted

    Raise:
        SchemaNotReadyException: if the schema is None
        DataNotMatchException: if the data is in consist with the schema
    """
    if schema is None:
        raise SchemaNotReadyException(message="Schema shouldn't be None")
    if schema.auto_id:
        if isinstance(data, pandas.DataFrame):
            if schema.primary_field.name in data:
                if not data[schema.primary_field.name].isnull().all():
                    raise DataNotMatchException(message=f"Please don't provide data for auto_id primary field: {schema.primary_field.name}")
                data = data.drop(schema.primary_field.name, axis=1)

    infer_fields = parse_fields_from_data(data)
    tmp_fields = copy.deepcopy(schema.fields)

    for i, field in enumerate(schema.fields):
        if field.is_primary and field.auto_id:
            tmp_fields.pop(i)

    if len(infer_fields) != len(tmp_fields):

        i_name = [f.name for f in infer_fields]
        t_name = [f.name for f in tmp_fields]
        raise DataNotMatchException(message=f"The fields don't match with schema fields, expected: {t_name}, got {i_name}")

    for x, y in zip(infer_fields, tmp_fields):
        if x.dtype != y.dtype:
            raise DataNotMatchException(message=f"The data type of field {y.name} doesn't match, expected: {y.dtype.name}, got {x.dtype.name}")
        if isinstance(data, pandas.DataFrame) and x.name != y.name:
            raise DataNotMatchException(message=f"The name of field don't match, expected: {y.name}, got {x.name}")


def parse_fields_from_data(data: [List[List], pandas.DataFrame]) -> List[FieldSchema]:
    if not isinstance(data, (pandas.DataFrame, list)):
        raise DataTypeNotSupportException(message="The type of data should be list or pandas.DataFrame")

    if isinstance(data, pandas.DataFrame):
        return parse_fields_from_dataframe(data)

    for d in data:
        if not is_list_like(d):
            raise DataTypeNotSupportException(message="data should be a list of list")

    fields = [FieldSchema("", infer_dtype_bydata(d[0])) for d in data]
    return fields


def parse_fields_from_dataframe(df: pandas.DataFrame) -> List[FieldSchema]:
    d_types = list(df.dtypes)
    data_types = list(map(map_numpy_dtype_to_datatype, d_types))
    col_names = list(df.columns)

    column_params_map = {}

    if DataType.UNKNOWN in data_types:
        if len(df) == 0:
            raise CannotInferSchemaException(message=ExceptionsMessage.DataFrameInvalid)
        values = df.head(1).values[0]
        for i, dtype in enumerate(data_types):
            if dtype == DataType.UNKNOWN:
                new_dtype = infer_dtype_bydata(values[i])
                if new_dtype in (DataType.BINARY_VECTOR, DataType.FLOAT_VECTOR):
                    vector_type_params = {}
                    if new_dtype == DataType.BINARY_VECTOR:
                        vector_type_params['dim'] = len(values[i]) * 8
                    else:
                        vector_type_params['dim'] = len(values[i])
                    column_params_map[col_names[i]] = vector_type_params
                data_types[i] = new_dtype

    if DataType.UNKNOWN in data_types:
        raise CannotInferSchemaException(message=ExceptionsMessage.DataFrameInvalid)

    fields = []
    for name, dtype in zip(col_names, data_types):
        type_params = column_params_map.get(name, {})
        field_schema = FieldSchema(name, dtype, **type_params)
        fields.append(field_schema)

    return fields


def check_schema(schema):
    if schema is None:
        raise SchemaNotReadyException(message=ExceptionsMessage.NoSchema)
    if len(schema.fields) < 1:
        raise SchemaNotReadyException(message=ExceptionsMessage.EmptySchema)
    vector_fields = []
    for field in schema.fields:
        if field.dtype in (DataType.FLOAT_VECTOR, DataType.BINARY_VECTOR):
            vector_fields.append(field.name)
    if len(vector_fields) < 1:
        raise SchemaNotReadyException(message=ExceptionsMessage.NoVector)



# hello_milvus.py demonstrates the basic operations of PyMilvus, a Python SDK of Milvus.
# 1. connect to Milvus
# 2. create collection
# 3. insert data
# 4. create index
# 5. search, query, and hybrid search on entities
# 6. delete entities by PK
# 7. drop collection
import time

import numpy as np
from pymilvus import (
    connections,
    utility,
    FieldSchema, CollectionSchema, DataType,
    Collection,
)

fmt = "\n=== {:30} ===\n"
search_latency_fmt = "search latency = {:.4f}s"
num_entities, dim = 3000, 8

#################################################################################
# 1. connect to Milvus
# Add a new connection alias `default` for Milvus server in `localhost:19530`
# Actually the "default" alias is a buildin in PyMilvus.
# If the address of Milvus is the same as `localhost:19530`, you can omit all
# parameters and call the method as: `connections.connect()`.
#
# Note: the `using` parameter of the following methods is default to "default".
print(fmt.format("start connecting to Milvus"))
connections.connect("default", host="localhost", port="19530")

has = utility.has_collection("hello_milvus")
print(f"Does collection hello_milvus exist in Milvus: {has}")

#################################################################################
# 2. create collection
# We're going to create a collection with 3 fields.
# +-+------------+------------+------------------+------------------------------+
# | | field name | field type | other attributes |       field description      |
# +-+------------+------------+------------------+------------------------------+
# |1|    "pk"    |   VarChar  |  is_primary=True |      "primary field"         |
# | |            |            |   auto_id=False  |                              |
# +-+------------+------------+------------------+------------------------------+
# |2|  "random"  |    Double  |                  |      "a double field"        |
# +-+------------+------------+------------------+------------------------------+
# |3|"embeddings"| FloatVector|     dim=8        |  "float vector with dim 8"   |
# +-+------------+------------+------------------+------------------------------+
fields = [
    FieldSchema(name="pk", dtype=DataType.VARCHAR, is_primary=True, auto_id=False, max_length=100),
    FieldSchema(name="random", dtype=DataType.DOUBLE),
    FieldSchema(name="embeddings", dtype=DataType.FLOAT_VECTOR, dim=dim)
]

schema = CollectionSchema(fields, "hello_milvus is the simplest demo to introduce the APIs")

print(fmt.format("Create collection `hello_milvus`"))
hello_milvus = Collection("hello_milvus", schema, consistency_level="Strong")

################################################################################
# 3. insert data
# We are going to insert 3000 rows of data into `hello_milvus`
# Data to be inserted must be organized in fields.
#
# The insert() method returns:
# - either automatically generated primary keys by Milvus if auto_id=True in the schema;
# - or the existing primary key field from the entities if auto_id=False in the schema.

print(fmt.format("Start inserting entities"))
rng = np.random.default_rng(seed=19530)
entities = [
    # provide the pk field because `auto_id` is set to False
    [str(i) for i in range(num_entities)],
    rng.random(num_entities).tolist(),  # field random, only supports list
    rng.random((num_entities, dim)),    # field embeddings, supports numpy.ndarray and list
]

insert_result = hello_milvus.insert(entities)

hello_milvus.flush()
print(f"Number of entities in Milvus: {hello_milvus.num_entities}")  # check the num_entites

################################################################################
# 4. create index
# We are going to create an IVF_FLAT index for hello_milvus collection.
# create_index() can only be applied to `FloatVector` and `BinaryVector` fields.
print(fmt.format("Start Creating index IVF_FLAT"))
index = {
    "index_type": "IVF_FLAT",
    "metric_type": "L2",
    "params": {"nlist": 128},
}

hello_milvus.create_index("embeddings", index)

################################################################################
# 5. search, query, and hybrid search
# After data were inserted into Milvus and indexed, you can perform:
# - search based on vector similarity
# - query based on scalar filtering(boolean, int, etc.)
# - hybrid search based on vector similarity and scalar filtering.
#

# Before conducting a search or a query, you need to load the data in `hello_milvus` into memory.
print(fmt.format("Start loading"))
hello_milvus.load()

# -----------------------------------------------------------------------------
# search based on vector similarity
print(fmt.format("Start searching based on vector similarity"))
vectors_to_search = entities[-1][-2:]
search_params = {
    "metric_type": "L2",
    "params": {"nprobe": 10},
}

start_time = time.time()
result = hello_milvus.search(vectors_to_search, "embeddings", search_params, limit=3, output_fields=["random"])
end_time = time.time()

for hits in result:
    for hit in hits:
        print(f"hit: {hit}, random field: {hit.entity.get('random')}")
print(search_latency_fmt.format(end_time - start_time))

# -----------------------------------------------------------------------------
# query based on scalar filtering(boolean, int, etc.)
print(fmt.format("Start querying with `random > 0.5`"))

start_time = time.time()
result = hello_milvus.query(expr="random > 0.5", output_fields=["random", "embeddings"])
end_time = time.time()

print(f"query result:\n-{result[0]}")
print(search_latency_fmt.format(end_time - start_time))

# -----------------------------------------------------------------------------
# pagination
r1 = hello_milvus.query(expr="random > 0.5", limit=4, output_fields=["random"])
r2 = hello_milvus.query(expr="random > 0.5", offset=1, limit=3, output_fields=["random"])
print(f"query pagination(limit=4):\n\t{r1}")
print(f"query pagination(offset=1, limit=3):\n\t{r2}")


# -----------------------------------------------------------------------------
# hybrid search
print(fmt.format("Start hybrid searching with `random > 0.5`"))

start_time = time.time()
result = hello_milvus.search(vectors_to_search, "embeddings", search_params, limit=3, expr="random > 0.5", output_fields=["random"])
end_time = time.time()

for hits in result:
    for hit in hits:
        print(f"hit: {hit}, random field: {hit.entity.get('random')}")
print(search_latency_fmt.format(end_time - start_time))

###############################################################################
# 6. delete entities by PK
# You can delete entities by their PK values using boolean expressions.
ids = insert_result.primary_keys

expr = f'pk in ["{ids[0]}" , "{ids[1]}"]'
print(fmt.format(f"Start deleting with expr `{expr}`"))

result = hello_milvus.query(expr=expr, output_fields=["random", "embeddings"])
print(f"query before delete by expr=`{expr}` -> result: \n-{result[0]}\n-{result[1]}\n")

hello_milvus.delete(expr)

result = hello_milvus.query(expr=expr, output_fields=["random", "embeddings"])
print(f"query after delete by expr=`{expr}` -> result: {result}\n")


###############################################################################
# 7. drop collection
# Finally, drop the hello_milvus collection
print(fmt.format("Drop collection `hello_milvus`"))
utility.drop_collection("hello_milvus")



import datetime
import threading
import pytest

from pymilvus.client import ts_utils


class TestTsUtils:
    def test_singleton(self):
        ins1 = ts_utils._get_gts_dict()
        ins2 = ts_utils._get_gts_dict()
        assert id(ins1) == id(ins2)

    def test_singleton_mutiple_thread(self):
        ins = ts_utils._get_gts_dict()

        def _f():
            g = ts_utils._get_gts_dict()
            assert id(g) == id(ins)

        t1 = threading.Thread(target=_f)
        t2 = threading.Thread(target=_f)

        t1.start()
        t2.start()

        t1.join()
        t2.join()

    def test_update_and_get(self):
        # test lru later if necessary.

        ins = ts_utils._get_gts_dict()
        assert ins.get(1) == 0

        ins.update("coll1", -1)
        assert ins.get("coll1") == 0

        ins.update("coll1", 2)
        assert ins.get("coll1") == 2

        ins.update("coll2", 100)
        assert ins.get("coll2") == 100

    def test_get_current_bounded_ts(self):
        ts = ts_utils.get_bounded_ts()
        assert ts == ts_utils.BOUNDED_TS

    def test_get_eventually_ts(self):
        ts = ts_utils.get_eventually_ts()
        assert ts == ts_utils.EVENTUALLY_TS





# Copyright (C) 2019-2020 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under the License.

from pymilvus import DataType, DEFAULT_RESOURCE_GROUP
from pymilvus.exceptions import InvalidConsistencyLevel
from pymilvus.client.types import (
    get_consistency_level, Shard, Group, Replica, ConsistencyLevel
)

from pymilvus.grpc_gen import common_pb2

import pytest
import pandas as pd
import numpy as np


@pytest.mark.xfail
class TestTypes:
    def test_map_numpy_dtype_to_datatype(self):
        data1 = {
            'double': [2.0],
            'float32': [np.float32(1.0)],
            'double2': [np.float64(1.0)],
            'int8': [np.int8(1)],
            'int16': [2],
            'int32': [4],
            'int64': [8],
            'bool': [True],
            'float_vec': [np.array([1.1, 1.2])],
        }

        df = pd.DataFrame(data1)

        wants1 = [
            DataType.DOUBLE,
            DataType.DOUBLE,
            DataType.DOUBLE,
            DataType.INT64,
            DataType.INT64,
            DataType.INT64,
            DataType.INT64,
            DataType.BOOL,
            DataType.UNKNOWN,
        ]

        ret1 = [map_numpy_dtype_to_datatype(x) for x in df.dtypes]
        assert ret1 == wants1

        df2 = pd.DataFrame(data=[1, 2, 3], columns=['a'],
                           dtype=np.int8)
        assert DataType.INT8 == map_numpy_dtype_to_datatype(df2.dtypes[0])

        df2 = pd.DataFrame(data=[1, 2, 3], columns=['a'],
                           dtype=np.int16)
        assert DataType.INT16 == map_numpy_dtype_to_datatype(df2.dtypes[0])

        df2 = pd.DataFrame(data=[1, 2, 3], columns=['a'],
                           dtype=np.int32)
        assert DataType.INT32 == map_numpy_dtype_to_datatype(df2.dtypes[0])

        df2 = pd.DataFrame(data=[1, 2, 3], columns=['a'],
                           dtype=np.int64)
        assert DataType.INT64 == map_numpy_dtype_to_datatype(df2.dtypes[0])

    def test_infer_dtype_bydata(self):
        data1 = [
            [1],
            [True],
            [1.0, 2.0],
            ["abc"],
            bytes("abc", encoding='ascii'),
            1,
            True,
            "abc",
            np.int8(1),
            np.int16(1),
            [np.int8(1)]
        ]

        wants = [
            DataType.FLOAT_VECTOR,
            DataType.UNKNOWN,
            DataType.FLOAT_VECTOR,
            DataType.UNKNOWN,
            DataType.BINARY_VECTOR,
            DataType.INT64,
            DataType.BOOL,
            DataType.STRING,
            DataType.INT8,
            DataType.INT16,
            DataType.FLOAT_VECTOR,
        ]

        actual = []
        for d in data1:
            actual.append(infer_dtype_bydata(d))

        assert actual == wants


class TestConsistencyLevel:
    def test_consistency_level_int(self):
        for v in ConsistencyLevel.values():
            assert v == get_consistency_level(v)

    def test_consistency_level_str(self):
        for k in ConsistencyLevel.keys():
            assert ConsistencyLevel.Value(k) == get_consistency_level(k)

    @pytest.mark.parametrize("invalid", [6, 100, "not supported", "中文", 1.0])
    def test_consistency_level_invalid(self, invalid):
        with pytest.raises(InvalidConsistencyLevel):
            get_consistency_level(invalid)


class TestReplica:
    def test_shard(self):
        s = Shard("channel-1", (1, 2, 3), 1)
        assert s.channel_name == "channel-1"
        assert s.shard_nodes == {1, 2, 3}
        assert s.shard_leader == 1
        print(s)

        g = Group(2, [s], [1, 2, 3], DEFAULT_RESOURCE_GROUP, {})
        assert g.id == 2
        assert g.shards == [s]
        assert g.group_nodes == (1, 2, 3)

        print(g)

        replica = Replica([g, g])
        assert replica.groups == [g, g]
        print(replica)

    def test_shard_dup_nodeIDs(self):
        s = Shard("channel-1", (1, 1, 1), 1)
        assert s.channel_name == "channel-1"
        assert s.shard_nodes == {1,}
        assert s.shard_leader == 1
        print(s)



class MockMutationResult:
    def __init__(self):
        self._primary_keys = []
        self._insert_cnt = 0
        self._delete_cnt = 0
        self._upsert_cnt = 0
        self._timestamp = 0

    @property
    def primary_keys(self):
        return self._primary_keys

    @property
    def insert_count(self):
        return self._insert_cnt

    @property
    def delete_count(self):
        return self._delete_cnt

    @property
    def upsert_count(self):
        return self._upsert_cnt

    @property
    def timestamp(self):
        return self._timestamp



import pytest
from pymilvus import Collection, Index, connections

from utils import (
    gen_index_name,
    gen_field_name,
    gen_collection_name,
    gen_schema,
    gen_index,
    gen_simple_index,
)


@pytest.mark.skip("Connect with the real server")
class TestIndex:
    @pytest.fixture(scope="function")
    def name(self):
        return gen_index_name()

    @pytest.fixture(scope="function")
    def field_name(self):
        return gen_field_name()

    @pytest.fixture(scope="function")
    def collection_name(self):
        return gen_collection_name()

    @pytest.fixture(scope="function")
    def schema(self):
        return gen_schema()

    @pytest.fixture(scope="function")
    def index_param(self):
        return gen_index()

    @pytest.fixture(
        scope="function",
        params=gen_simple_index()
    )
    def get_simple_index(self, request):
        return request.param

    @pytest.fixture(scope="function")
    def index(self, name, field_name, collection_name, schema, get_simple_index):
        connections.connect()
        collection = Collection(collection_name, schema=schema)
        return Index(collection, field_name, get_simple_index)

    def test_params(self, index, get_simple_index):
        assert index.params == get_simple_index

    def test_collection_name(self, index, collection_name):
        assert index.collection_name == collection_name

    def test_field_name(self, index, field_name):
        assert index.field_name == field_name

    def test_drop(self, index):
        index.drop()



# Copyright (C) 2019-2020 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under the License.

from pymilvus import Collection, connections, FieldSchema, CollectionSchema, DataType, utility
from pymilvus.client.types import LoadState

import random
import numpy as np
from sklearn import preprocessing
import string

default_dim = 128
default_nb = 3000
default_float_vec_field_name = "float_vector"
default_binary_vec_field_name = "binary_vector"

all_index_types = [
    "FLAT",
    "IVF_FLAT",
    "IVF_SQ8",
    # "IVF_SQ8_HYBRID",
    "IVF_PQ",
    "HNSW",
    # "NSG",
    "ANNOY",
    "RHNSW_FLAT",
    "RHNSW_PQ",
    "RHNSW_SQ",
    "BIN_FLAT",
    "BIN_IVF_FLAT"
]

default_index_params = [
    {"nlist": 128},
    {"nlist": 128},
    {"nlist": 128},
    # {"nlist": 128},
    {"nlist": 128, "m": 16, "nbits": 8},
    {"M": 48, "efConstruction": 500},
    # {"search_length": 50, "out_degree": 40, "candidate_pool_size": 100, "knng": 50},
    {"n_trees": 50},
    {"M": 48, "efConstruction": 500},
    {"M": 48, "efConstruction": 500, "PQM": 64},
    {"M": 48, "efConstruction": 500},
    {"nlist": 128},
    {"nlist": 128}
]

default_index = {"index_type": "IVF_FLAT", "params": {"nlist": 128}, "metric_type": "L2"}
default_binary_index = {"index_type": "BIN_FLAT", "params": {"nlist": 1024}, "metric_type": "JACCARD"}


def gen_default_fields(description="test collection"):
    default_fields = [
        FieldSchema(name="int64", dtype=DataType.INT64, is_primary=True),
        FieldSchema(name="double", dtype=DataType.DOUBLE),
        FieldSchema(name=default_float_vec_field_name, dtype=DataType.FLOAT_VECTOR, dim=default_dim)
    ]
    default_schema = CollectionSchema(fields=default_fields, description=description)
    return default_schema


def gen_default_fields_with_primary_key_1():
    default_fields = [
        FieldSchema(name="int64", dtype=DataType.INT64, is_primary=True),
        FieldSchema(name="double", dtype=DataType.DOUBLE),
        FieldSchema(name=default_float_vec_field_name, dtype=DataType.FLOAT_VECTOR, dim=default_dim)
    ]
    default_schema = CollectionSchema(fields=default_fields, description="test collection")
    return default_schema


def gen_default_fields_with_primary_key_2():
    default_fields = [
        FieldSchema(name="int64", dtype=DataType.INT64),
        FieldSchema(name="double", dtype=DataType.DOUBLE),
        FieldSchema(name=default_float_vec_field_name, dtype=DataType.FLOAT_VECTOR, dim=default_dim)
    ]
    default_schema = CollectionSchema(fields=default_fields, description="test collection", primary_field="int64")
    return default_schema


def gen_binary_schema():
    binary_fields = [
        FieldSchema(name="int64", dtype=DataType.INT64, is_primary=True),
        FieldSchema(name="double", dtype=DataType.DOUBLE),
        FieldSchema(name=default_binary_vec_field_name, dtype=DataType.BINARY_VECTOR, dim=default_dim)
    ]
    default_schema = CollectionSchema(fields=binary_fields, description="test collection")
    return default_schema


def gen_float_vectors(num, dim, is_normal=True):
    vectors = [[random.random() for _ in range(dim)] for _ in range(num)]
    vectors = preprocessing.normalize(vectors, axis=1, norm='l2')
    return vectors.tolist()


def gen_float_data(nb, is_normal=False):
    vectors = gen_float_vectors(nb, default_dim, is_normal)
    entities = [
        [i for i in range(nb)],
        [float(i) for i in range(nb)],
        vectors
    ]
    return entities


def gen_dataframe(nb, is_normal=False):
    import pandas
    import numpy

    vectors = gen_float_vectors(nb, default_dim, is_normal)
    data = {
        "int64": [i for i in range(nb)],
        "float": numpy.array([i for i in range(nb)], dtype=numpy.float32),
        "float_vector": vectors
    }

    return pandas.DataFrame(data)


def gen_binary_vectors(num, dim):
    raw_vectors = []
    binary_vectors = []
    for i in range(num):
        raw_vector = [random.randint(0, 1) for i in range(dim)]
        raw_vectors.append(raw_vector)
        binary_vectors.append(bytes(np.packbits(raw_vector, axis=-1).tolist()))
    return raw_vectors, binary_vectors


def gen_binary_data(nb):
    raw_vectors, binary_vectors = gen_binary_vectors(nb, dim=default_dim)
    entities = [
        [i for i in range(nb)],
        [float(i) for i in range(nb)],
        binary_vectors
    ]
    return entities


def gen_unique_str(str_value=None):
    prefix = "".join(random.choice(string.ascii_letters + string.digits) for _ in range(8))
    return "collection_" + prefix if str_value is None else str_value + "_" + prefix


def binary_support():
    return ["BIN_FLAT", "BIN_IVF_FLAT"]


def gen_simple_index():
    index_params = []
    for i in range(len(all_index_types)):
        if all_index_types[i] in binary_support():
            continue
        dic = {"index_type": all_index_types[i], "metric_type": "L2"}
        dic.update({"params": default_index_params[i]})
        index_params.append(dic)
    return index_params


connections.connect(alias="default")


def test_create_collection():
    name = gen_unique_str()
    collection = Collection(name=name, schema=gen_default_fields())
    assert collection.is_empty is True
    assert collection.num_entities == 0
    assert utility.load_state(name) == LoadState.NotLoad
    return name


def test_exist_collection(name):
    assert utility.has_collection(name) is True
    collection = Collection(name)
    collection.drop()


def test_collection_only_name():
    name = gen_unique_str()
    Collection(name=name, schema=gen_default_fields())
    collection = Collection(name=name)
    data = gen_float_data(default_nb)
    collection.insert(data)
    collection.flush()
    collection.create_index(field_name=default_float_vec_field_name, index_params=default_index)
    collection.load()
    assert collection.is_empty is False
    assert collection.num_entities == default_nb
    assert utility.load_state(name) == LoadState.Loaded
    collection.drop()


def test_collection_with_dataframe():
    data = gen_dataframe(default_nb)
    collection, _ = Collection.construct_from_dataframe(name=gen_unique_str(), dataframe=data, primary_field="int64")
    collection.flush()
    collection.create_index(field_name=default_float_vec_field_name, index_params=default_index)
    collection.load()
    assert collection.is_empty is False
    assert collection.num_entities == default_nb
    collection.drop()


def test_create_index_float_vector():
    data = gen_float_data(default_nb)
    collection = Collection(name=gen_unique_str(), data=data, schema=gen_default_fields())
    for index_param in gen_simple_index():
        collection.create_index(field_name=default_float_vec_field_name, index_params=index_param)
        assert len(collection.indexes) != 0
        collection.drop_index()
    collection.drop()


def test_create_index_binary_vector():
    collection = Collection(name=gen_unique_str(), schema=gen_binary_schema())
    data = gen_binary_data(default_nb)
    collection.insert(data)
    collection.create_index(field_name=default_binary_vec_field_name, index_params=default_binary_index)
    assert len(collection.indexes) != 0
    collection.drop()


def test_specify_primary_key():
    data = gen_float_data(default_nb)
    collection = Collection(name=gen_unique_str(), data=data, schema=gen_default_fields_with_primary_key_1())
    for index_param in gen_simple_index():
        collection.create_index(field_name=default_float_vec_field_name, index_params=index_param)
        assert len(collection.indexes) != 0
        collection.drop_index()
    collection.drop()

    collection2 = Collection(name=gen_unique_str(), data=data, schema=gen_default_fields_with_primary_key_2())
    for index_param in gen_simple_index():
        collection2.create_index(field_name=default_float_vec_field_name, index_params=index_param)
        assert len(collection2.indexes) != 0
        collection2.drop_index()
    collection2.drop()


def test_alias():
    """ Test alias follows the following steps
    1. Prepare tests
    2. Create collection_A and create an alias `latest_collection`
    3. Create collection_B and shift alias `latest_collection` to collection_B
    4. Drop alias `latest_collection`
    5. Clear up tests

    """

    name_A, name_B = "collection_A", "collection_B"

    def setup() -> (Collection, Collection):
        collection_A = Collection(name=name_A, schema=gen_default_fields("collection A"))
        collection_B = Collection(name=name_B, schema=gen_default_fields("collection B"))
        return collection_A, collection_B

    def teardown():
        if utility.has_collection(name_A):
            utility.drop_collection(name_A)
        if utility.has_collection(name_B):
            utility.drop_collection(name_B)

    def alias_cases():
        teardown()
        A, B = setup()

        latest_coll_alias = "latest_collection"

        utility.create_alias(A.name, latest_coll_alias)

        alias_collection = Collection(latest_coll_alias)
        assert alias_collection.description == A.description

        utility.alter_alias(B.name, latest_coll_alias)

        alias_collection = Collection(latest_coll_alias)
        assert alias_collection.description == B.description

        utility.drop_alias(latest_coll_alias)
        try:
            alias_collection = Collection(latest_coll_alias)
        except BaseException as e:
            print(f" - Alias [{latest_coll_alias}] dropped, cannot get collection from it. Error msg: {e}")
        finally:
            teardown()

    alias_cases()


def test_rename_collection():
    connections.connect(alias="default")
    schema = CollectionSchema(fields=[
        FieldSchema("int64", DataType.INT64, description="int64", is_primary=True),
        FieldSchema("float_vector", DataType.FLOAT_VECTOR, is_primary=False, dim=128),
    ])

    old_collection = "old_collection"
    new_collection = "new_collection"
    Collection(old_collection, schema=schema)

    print("\nlist collections:")
    print(utility.list_collections())
    assert utility.has_collection(old_collection)

    utility.rename_collection(old_collection, new_collection)
    assert utility.has_collection(new_collection)
    assert not utility.has_collection(old_collection)


if __name__ == "__main__":
    print("test collection and get an existing collection")
    name = test_create_collection()
    print("test an existing collection")
    test_exist_collection(name)
    print("test collection only name")
    test_collection_only_name()
    print("test collection with dataframe")
    test_collection_with_dataframe()
    print("test collection index float vector")
    test_create_index_float_vector()
    print("test collection binary vector")
    test_create_index_binary_vector()
    print("test collection specify primary key")
    test_specify_primary_key()
    print("test alias")
    test_alias()
    print("test rename collection")
    test_rename_collection()
    print("test end")



import copy
import base64
from typing import Dict, Iterable, Union

import ujson

from . import blob
from . import entity_helper
from .check import check_pass_param, is_legal_collection_properties
from .types import DataType, PlaceholderType, get_consistency_level
from .constants import DEFAULT_CONSISTENCY_LEVEL
from ..exceptions import ParamError, DataNotMatchException, ExceptionsMessage
from ..orm.schema import CollectionSchema

from ..grpc_gen import common_pb2 as common_types
from ..grpc_gen import schema_pb2 as schema_types
from ..grpc_gen import milvus_pb2 as milvus_types


class Prepare:
    @classmethod
    def create_collection_request(cls, collection_name: str, fields: Union[Dict[str, Iterable], CollectionSchema],
                                  shards_num=2, **kwargs) -> milvus_types.CreateCollectionRequest:
        """
        :type fields: Union(Dict[str, Iterable], CollectionSchema)
        :param fields: (Required)

            `{"fields": [
                    {"name": "A", "type": DataType.INT32}
                    {"name": "B", "type": DataType.INT64, "auto_id": True, "is_primary": True},
                    {"name": "C", "type": DataType.FLOAT},
                    {"name": "Vec", "type": DataType.FLOAT_VECTOR, "params": {"dim": 128}}]
            }`

        :return: milvus_types.CreateCollectionRequest
        """
        if isinstance(fields, CollectionSchema):
            schema = cls.get_schema_from_collection_schema(collection_name, fields, shards_num, **kwargs)
        else:
            schema = cls.get_schema(collection_name, fields, shards_num, **kwargs)

        consistency_level = get_consistency_level(kwargs.get("consistency_level", DEFAULT_CONSISTENCY_LEVEL))

        req = milvus_types.CreateCollectionRequest(collection_name=collection_name,
                                                    schema=bytes(schema.SerializeToString()),
                                                    shards_num=shards_num,
                                                    consistency_level=consistency_level)

        properties = kwargs.get("properties")
        if is_legal_collection_properties(properties):
            properties = [common_types.KeyValuePair(key=str(k), value=str(v)) for k, v in properties.items()]
            req.properties.extend(properties)

        return req

    @classmethod
    def get_schema_from_collection_schema(cls, collection_name: str, fields: CollectionSchema, shards_num=2, **kwargs) -> milvus_types.CreateCollectionRequest:
        coll_description = fields.description
        if not isinstance(coll_description, (str, bytes)):
            raise ParamError(message=f"description [{coll_description}] has type {type(coll_description).__name__},  but expected one of: bytes, str")

        schema = schema_types.CollectionSchema(name=collection_name,
                                               autoID=fields.auto_id,
                                               description=coll_description)
        for f in fields.fields:
            field_schema = schema_types.FieldSchema(name=f.name,
                                                    data_type=f.dtype,
                                                    description=f.description,
                                                    is_primary_key=f.is_primary,
                                                    autoID=f.auto_id)
            for k, v in f.params.items():
                kv_pair = common_types.KeyValuePair(key=str(k), value=str(v))
                field_schema.type_params.append(kv_pair)

            schema.fields.append(field_schema)
        return schema

    @classmethod
    def get_schema(cls, collection_name: str, fields: Dict[str, Iterable], shards_num=2, **kwargs) -> schema_types.CollectionSchema:
        if not isinstance(fields, dict):
            raise ParamError(message="Param fields must be a dict")

        all_fields = fields.get("fields")
        if all_fields is None:
            raise ParamError(message="Param fields must contain key 'fields'")
        if len(all_fields) == 0:
            raise ParamError(message="Param fields value cannot be empty")

        schema = schema_types.CollectionSchema(name=collection_name,
                                               autoID=False,
                                               description=fields.get('description', ''))

        primary_field = None
        auto_id_field = None
        for field in all_fields:
            field_name = field.get('name')
            if field_name is None:
                raise ParamError(message="You should specify the name of field!")

            data_type = field.get('type')
            if data_type is None:
                raise ParamError(message="You should specify the data type of field!")
            if not isinstance(data_type, (int, DataType)):
                raise ParamError(message="Field type must be of DataType!")

            is_primary = field.get("is_primary", False)
            if not isinstance(is_primary, bool):
                raise ParamError(message="is_primary must be boolean")
            if is_primary:
                if primary_field is not None:
                    raise ParamError(message="A collection should only have one primary field")
                if DataType(data_type) not in [DataType.INT64, DataType.VARCHAR]:
                    raise ParamError(message="int64 and varChar are the only supported types of primary key")
                primary_field = field_name

            auto_id = field.get('auto_id', False)
            if not isinstance(auto_id, bool):
                raise ParamError(message="auto_id must be boolean")
            if auto_id:
                if auto_id_field is not None:
                    raise ParamError(message="A collection should only have one autoID field")
                if DataType(data_type) != DataType.INT64:
                    raise ParamError(message="int64 is the only supported type of automatic generated id")
                auto_id_field = field_name

            field_schema = schema_types.FieldSchema(name=field_name,
                                                    data_type=data_type,
                                                    description=field.get('description', ''),
                                                    is_primary_key=is_primary,
                                                    autoID=auto_id)

            type_params = field.get('params', {})
            if not isinstance(type_params, dict):
                raise ParamError(message="params should be dictionary type")
            kvs = [common_types.KeyValuePair(key=str(k), value=str(v)) for k, v in type_params.items()]
            field_schema.type_params.extend(kvs)

            schema.fields.append(field_schema)
        return schema

    @classmethod
    def drop_collection_request(cls, collection_name):
        return milvus_types.DropCollectionRequest(collection_name=collection_name)

    @classmethod
    # TODO remove
    def has_collection_request(cls, collection_name):
        return milvus_types.HasCollectionRequest(collection_name=collection_name)

    @classmethod
    def describe_collection_request(cls, collection_name):
        return milvus_types.DescribeCollectionRequest(collection_name=collection_name)

    @classmethod
    def alter_collection_request(cls, collection_name, properties):
        kvs = []
        for k in properties:
            kv = common_types.KeyValuePair(key=k, value=str(properties[k]))
            kvs.append(kv)

        return milvus_types.AlterCollectionRequest(collection_name=collection_name, properties=kvs)

    @classmethod
    def collection_stats_request(cls, collection_name):
        return milvus_types.CollectionStatsRequest(collection_name=collection_name)

    @classmethod
    def show_collections_request(cls, collection_names=None):
        req = milvus_types.ShowCollectionsRequest()
        if collection_names:
            if not isinstance(collection_names, (list,)):
                raise ParamError(message=f"collection_names must be a list of strings, but got: {collection_names}")
            for collection_name in collection_names:
                check_pass_param(collection_name=collection_name)
            req.collection_names.extend(collection_names)
            req.type = milvus_types.ShowType.InMemory
        return req

    @classmethod
    def rename_collections_request(cls, old_name=None, new_name=None):
        return milvus_types.RenameCollectionRequest(oldName=old_name, newName=new_name)

    @classmethod
    def create_partition_request(cls, collection_name, partition_name):
        return milvus_types.CreatePartitionRequest(collection_name=collection_name, partition_name=partition_name)

    @classmethod
    def drop_partition_request(cls, collection_name, partition_name):
        return milvus_types.DropPartitionRequest(collection_name=collection_name, partition_name=partition_name)

    @classmethod
    def has_partition_request(cls, collection_name, partition_name):
        return milvus_types.HasPartitionRequest(collection_name=collection_name, partition_name=partition_name)

    @classmethod
    def partition_stats_request(cls, collection_name, partition_name):
        return milvus_types.PartitionStatsRequest(collection_name=collection_name, partition_name=partition_name)

    @classmethod
    def show_partitions_request(cls, collection_name, partition_names=None, type_in_memory=False):
        check_pass_param(collection_name=collection_name, partition_name_array=partition_names)
        req = milvus_types.ShowPartitionsRequest(collection_name=collection_name)
        if partition_names:
            if not isinstance(partition_names, (list,)):
                raise ParamError(message=f"partition_names must be a list of strings, but got: {partition_names}")
            for partition_name in partition_names:
                check_pass_param(partition_name=partition_name)
            req.partition_names.extend(partition_names)
        if type_in_memory is False:
            req.type = milvus_types.ShowType.All
        else:
            req.type = milvus_types.ShowType.InMemory
        return req

    @classmethod
    def get_loading_progress(cls, collection_name, partition_names=None):
        check_pass_param(collection_name=collection_name, partition_name_array=partition_names)
        req = milvus_types.GetLoadingProgressRequest(collection_name=collection_name)
        if partition_names:
            req.partition_names.extend(partition_names)
        return req

    @classmethod
    def get_load_state(cls, collection_name, partition_names=None):
        check_pass_param(collection_name=collection_name, partition_name_array=partition_names)
        req = milvus_types.GetLoadStateRequest(collection_name=collection_name)
        if partition_names:
            req.partition_names.extend(partition_names)
        return req

    @classmethod
    def empty(cls):
        raise DeprecationWarning("no empty request later")
        # return common_types.Empty()

    @classmethod
    def register_link_request(cls):
        return milvus_types.RegisterLinkRequest()

    @classmethod
    def partition_name(cls, collection_name, partition_name):
        if not isinstance(collection_name, str):
            raise ParamError(message="collection_name must be of str type")
        if not isinstance(partition_name, str):
            raise ParamError(message="partition_name must be of str type")
        return milvus_types.PartitionName(collection_name=collection_name,
                                          tag=partition_name)

    @classmethod
    def batch_insert_param(cls, collection_name, entities, partition_name, fields_info=None, **kwargs):
        # insert_request.hash_keys won't be filled in client. It will be filled in proxy.

        tag = partition_name or "_default" # should here?
        insert_request = milvus_types.InsertRequest(collection_name=collection_name, partition_name=tag)

        for entity in entities:
            if not entity.get("name", None) or not entity.get("values", None) or not entity.get("type", None):
                raise ParamError(message="Missing param in entities, a field must have type, name and values")
        if not fields_info:
            raise ParamError(message="Missing collection meta to validate entities")

        location, primary_key_loc, auto_id_loc = {}, None, None
        for i, field in enumerate(fields_info):
            if field.get("is_primary", False):
                primary_key_loc = i

            if field.get("auto_id", False):
                auto_id_loc = i
                continue

            match_flag = False
            field_name = field["name"]
            field_type = field["type"]

            for entity in entities:
                entity_name, entity_type = entity["name"], entity["type"]

                if field_name == entity_name:
                    if field_type != entity_type:
                        raise ParamError(message=f"Collection field type is {field_type}"
                                         f", but entities field type is {entity_type}")

                    entity_dim, field_dim = 0, 0
                    if entity_type in [DataType.FLOAT_VECTOR, DataType.BINARY_VECTOR]:
                        field_dim = field["params"]["dim"]
                        entity_dim = len(entity["values"][0])

                    if entity_type in [DataType.FLOAT_VECTOR, ] and entity_dim != field_dim:
                        raise ParamError(message=f"Collection field dim is {field_dim}"
                                         f", but entities field dim is {entity_dim}")

                    if entity_type in [DataType.BINARY_VECTOR, ] and entity_dim * 8 != field_dim:
                        raise ParamError(message=f"Collection field dim is {field_dim}"
                                         f", but entities field dim is {entity_dim * 8}")

                    location[field["name"]] = i
                    match_flag = True
                    break

            if not match_flag:
                raise ParamError(message=f"Field {field['name']} don't match in entities")

        # though impossible from sdk
        if primary_key_loc is None:
            raise ParamError(message="primary key not found")

        if auto_id_loc is None and len(entities) != len(fields_info):
            raise ParamError(message=f"number of fields: {len(fields_info)}, number of entities: {len(entities)}")

        if auto_id_loc is not None and len(entities) + 1 != len(fields_info):
            raise ParamError(message=f"number of fields: {len(fields_info)}, number of entities: {len(entities)}")

        row_num = 0
        try:
            for entity in entities:
                current = len(entity.get("values"))
                if row_num not in (0, current):
                    raise ParamError(message="row num misaligned current[{current}]!= previous[{row_num}]")
                row_num = current
                field_data = entity_helper.entity_to_field_data(entity, fields_info[location[entity.get("name")]])
                insert_request.fields_data.append(field_data)
        except (TypeError, ValueError) as e:
            raise DataNotMatchException(message=ExceptionsMessage.DataTypeInconsistent) from e

        insert_request.num_rows = row_num

        return insert_request

    @classmethod
    def delete_request(cls, collection_name, partition_name, expr):
        def check_str(instr, prefix):
            if instr is None:
                raise ParamError(message=f"{prefix} cannot be None")
            if not isinstance(instr, str):
                raise ParamError(message=f"{prefix} value {instr} is illegal")
            if instr == "":
                raise ParamError(message=f"{prefix} cannot be empty")

        check_str(collection_name, "collection_name")
        if partition_name is not None and partition_name != "":
            check_str(partition_name, "partition_name")
        check_str(expr, "expr")

        request = milvus_types.DeleteRequest(collection_name=collection_name, expr=expr, partition_name=partition_name)
        return request

    @classmethod
    def _prepare_placeholders(cls, vectors, nq, tag, pl_type, is_binary, dimension=0):
        pl = common_types.PlaceholderValue(tag=tag)
        pl.type = pl_type
        for i in range(0, nq):
            if is_binary:
                if len(vectors[i]) * 8 != dimension:
                    raise ParamError(message=f"The dimension of query entities[{vectors[i]*8}] is different from schema [{dimension}]")
                pl.values.append(blob.vectorBinaryToBytes(vectors[i]))
            else:
                if len(vectors[i]) != dimension:
                    raise ParamError(message=f"The dimension of query entities[{vectors[i]*8}] is different from schema [{dimension}]")
                pl.values.append(blob.vectorFloatToBytes(vectors[i]))
        return pl

    @classmethod
    def search_request(cls, collection_name, query_entities, partition_names=None, fields=None, round_decimal=-1, **kwargs):
        schema = kwargs.get("schema", None)
        fields_schema = schema.get("fields", None)  # list
        fields_name_locs = {fields_schema[loc]["name"]: loc
                            for loc in range(len(fields_schema))}

        if not isinstance(query_entities, (dict,)):
            raise ParamError(message="Invalid query format. 'query_entities' must be a dict")

        if fields is not None and not isinstance(fields, (list,)):
            raise ParamError(message="Invalid query format. 'fields' must be a list")

        request = milvus_types.SearchRequest(
            collection_name=collection_name,
            partition_names=partition_names,
            output_fields=fields,
            guarantee_timestamp=kwargs.get("guarantee_timestamp", 0),
        )

        duplicated_entities = copy.deepcopy(query_entities)
        vector_placeholders = {}
        vector_names = {}

        def extract_vectors_param(param, placeholders, names, round_decimal):
            if not isinstance(param, (dict, list)):
                return

            if isinstance(param, dict):
                if "vector" in param:
                    # TODO: Here may not replace ph
                    ph = "$" + str(len(placeholders))

                    for pk, pv in param["vector"].items():
                        if "query" not in pv:
                            raise ParamError(message="param vector must contain 'query'")
                        placeholders[ph] = pv["query"]
                        names[ph] = pk
                        param["vector"][pk]["query"] = ph
                        param["vector"][pk]["round_decimal"] = round_decimal
                    return

                for _, v in param.items():
                    extract_vectors_param(v, placeholders, names, round_decimal)

            if isinstance(param, list):
                for item in param:
                    extract_vectors_param(item, placeholders, names, round_decimal)

        extract_vectors_param(duplicated_entities, vector_placeholders, vector_names, round_decimal)
        request.dsl = ujson.dumps(duplicated_entities)

        plg = common_types.PlaceholderGroup()
        for tag, vectors in vector_placeholders.items():
            if len(vectors) <= 0:
                continue
            pl = common_types.PlaceholderValue(tag=tag)

            fname = vector_names[tag]
            if fname not in fields_name_locs:
                raise ParamError(message=f"Field {fname} doesn't exist in schema")
            dimension = int(fields_schema[fields_name_locs[fname]]["params"].get("dim", 0))

            if isinstance(vectors[0], bytes):
                pl.type = PlaceholderType.BinaryVector
                for vector in vectors:
                    if dimension != len(vector) * 8:
                        raise ParamError(message="The dimension of query vector is different from schema")
                    pl.values.append(blob.vectorBinaryToBytes(vector))
            else:
                pl.type = PlaceholderType.FloatVector
                for vector in vectors:
                    if dimension != len(vector):
                        raise ParamError(message="The dimension of query vector is different from schema")
                    pl.values.append(blob.vectorFloatToBytes(vector))
            # vector_values_bytes = service_msg_types.VectorValues.SerializeToString(vector_values)

            plg.placeholders.append(pl)
        plg_str = common_types.PlaceholderGroup.SerializeToString(plg)
        request.placeholder_group = plg_str

        return request

    @classmethod
    def search_requests_with_expr(cls, collection_name, data, anns_field, param, limit, schema, expr=None,
                                  partition_names=None, output_fields=None, round_decimal=-1, **kwargs):
        # TODO Move this impl into server side
        fields_schema = schema.get("fields", None)  # list
        fields_name_locs = {fields_schema[loc]["name"]: loc for loc in range(len(fields_schema))}

        requests = []
        if len(data) <= 0:
            return requests

        if isinstance(data[0], bytes):
            is_binary = True
            pl_type = PlaceholderType.BinaryVector
        else:
            is_binary = False
            pl_type = PlaceholderType.FloatVector

        if anns_field not in fields_name_locs:
            raise ParamError(message=f"Field {anns_field} doesn't exist in schema")
        dimension = int(fields_schema[fields_name_locs[anns_field]]["params"].get("dim", 0))

        ignore_growing = param.get("ignore_growing",False) or kwargs.get("ignore_growing",False)
        params = param.get("params", {})
        if not isinstance(params, dict):
            raise ParamError(message=f"Search params must be a dict, got {type(params)}")
        search_params = {
            "anns_field": anns_field,
            "topk": limit,
            "metric_type": param.get("metric_type", "L2"),
            "params": params,
            "round_decimal": round_decimal,
            "offset": param.get("offset", 0),
            "ignore_growing": ignore_growing,
        }

        def dump(v):
            if isinstance(v, dict):
                return ujson.dumps(v)
            return str(v)

        nq = len(data)
        tag = "$0"
        pl = cls._prepare_placeholders(data, nq, tag, pl_type, is_binary, dimension)
        plg = common_types.PlaceholderGroup()
        plg.placeholders.append(pl)
        plg_str = common_types.PlaceholderGroup.SerializeToString(plg)
        request = milvus_types.SearchRequest(
            collection_name=collection_name,
            partition_names=partition_names,
            output_fields=output_fields,
            guarantee_timestamp=kwargs.get("guarantee_timestamp", 0),
            travel_timestamp=kwargs.get("travel_timestamp", 0),
            nq=nq,
        )
        request.placeholder_group = plg_str

        request.dsl_type = common_types.DslType.BoolExprV1
        if expr is not None:
            request.dsl = expr
        request.search_params.extend([common_types.KeyValuePair(key=str(key), value=dump(value))
                                      for key, value in search_params.items()])

        requests.append(request)
        return requests

    @classmethod
    def create_alias_request(cls, collection_name, alias):
        return milvus_types.CreateAliasRequest(collection_name=collection_name, alias=alias)

    @classmethod
    def drop_alias_request(cls, alias):
        return milvus_types.DropAliasRequest(alias=alias)

    @classmethod
    def alter_alias_request(cls, collection_name, alias):
        return milvus_types.AlterAliasRequest(collection_name=collection_name, alias=alias)

    @classmethod
    def create_index_request(cls, collection_name, field_name, params, **kwargs):
        index_params = milvus_types.CreateIndexRequest(collection_name=collection_name, field_name=field_name,
                                                       index_name=kwargs.get("index_name", ""))

        # index_params.collection_name = collection_name
        # index_params.field_name = field_name

        def dump(tv):
            if isinstance(tv, dict):
                import json
                return json.dumps(tv)
            return str(tv)

        if isinstance(params, dict):
            for tk, tv in params.items():
                if tk == "dim":
                    if not tv or not isinstance(tv, int):
                        raise ParamError(message="dim must be of int!")
                kv_pair = common_types.KeyValuePair(key=str(tk), value=dump(tv))
                index_params.extra_params.append(kv_pair)

        return index_params

    @classmethod
    def describe_index_request(cls, collection_name, index_name):
        return milvus_types.DescribeIndexRequest(collection_name=collection_name, index_name=index_name)

    @classmethod
    def get_index_build_progress(cls, collection_name: str, index_name: str):
        return milvus_types.GetIndexBuildProgressRequest(collection_name=collection_name, index_name=index_name)

    @classmethod
    def get_index_state_request(cls, collection_name: str, index_name: str):
        return milvus_types.GetIndexStateRequest(collection_name=collection_name, index_name=index_name)

    @classmethod
    def load_collection(cls, db_name, collection_name, replica_number, refresh, resource_groups):
        return milvus_types.LoadCollectionRequest(db_name=db_name, collection_name=collection_name,
                                                  replica_number=replica_number, refresh=refresh,
                                                  resource_groups=resource_groups)

    @classmethod
    def release_collection(cls, db_name, collection_name):
        return milvus_types.ReleaseCollectionRequest(db_name=db_name, collection_name=collection_name)

    @classmethod
    def load_partitions(cls, db_name, collection_name, partition_names, replica_number, refresh, resource_groups):
        return milvus_types.LoadPartitionsRequest(db_name=db_name, collection_name=collection_name,
                                                  partition_names=partition_names,
                                                  replica_number=replica_number,
                                                  refresh=refresh,
                                                  resource_groups=resource_groups)

    @classmethod
    def release_partitions(cls, db_name, collection_name, partition_names):
        return milvus_types.ReleasePartitionsRequest(db_name=db_name, collection_name=collection_name,
                                                     partition_names=partition_names)

    @classmethod
    def get_collection_stats_request(cls, collection_name):
        return milvus_types.GetCollectionStatisticsRequest(collection_name=collection_name)

    @classmethod
    def get_persistent_segment_info_request(cls, collection_name):
        return milvus_types.GetPersistentSegmentInfoRequest(collectionName=collection_name)

    @classmethod
    def get_flush_state_request(cls, segment_ids):
        return milvus_types.GetFlushStateRequest(segmentIDs=segment_ids)

    @classmethod
    def get_query_segment_info_request(cls, collection_name):
        return milvus_types.GetQuerySegmentInfoRequest(collectionName=collection_name)

    @classmethod
    def flush_param(cls, collection_names):
        return milvus_types.FlushRequest(collection_names=collection_names)

    @classmethod
    def drop_index_request(cls, collection_name, field_name, index_name):
        return milvus_types.DropIndexRequest(db_name="", collection_name=collection_name, field_name=field_name,
                                             index_name=index_name)

    @classmethod
    def get_partition_stats_request(cls, collection_name, partition_name):
        return milvus_types.GetPartitionStatisticsRequest(db_name="", collection_name=collection_name,
                                                          partition_name=partition_name)

    @classmethod
    def dummy_request(cls, request_type):
        return milvus_types.DummyRequest(request_type=request_type)

    @classmethod
    def retrieve_request(cls, collection_name, ids, output_fields, partition_names):
        ids = schema_types.IDs(int_id=schema_types.LongArray(data=ids))
        return milvus_types.RetrieveRequest(db_name="",
                                            collection_name=collection_name,
                                            ids=ids,
                                            output_fields=output_fields,
                                            partition_names=partition_names)

    @classmethod
    def query_request(cls, collection_name, expr, output_fields, partition_names, **kwargs):
        req = milvus_types.QueryRequest(db_name="",
                                        collection_name=collection_name,
                                        expr=expr,
                                        output_fields=output_fields,
                                        partition_names=partition_names,
                                        guarantee_timestamp=kwargs.get("guarantee_timestamp", 0),
                                        travel_timestamp=kwargs.get("travel_timestamp", 0),
                                        )

        limit = kwargs.get("limit", None)
        if limit is not None:
            req.query_params.append(common_types.KeyValuePair(key="limit", value=str(limit)))

        offset = kwargs.get("offset", None)
        if offset is not None:
            req.query_params.append(common_types.KeyValuePair(key="offset", value=str(offset)))

        ignore_growing = kwargs.get("ignore_growing", False)
        req.query_params.append(common_types.KeyValuePair(key="ignore_growing", value=str(ignore_growing)))
        return req

    @classmethod
    def load_balance_request(cls, collection_name, src_node_id, dst_node_ids, sealed_segment_ids):
        request = milvus_types.LoadBalanceRequest(
            collectionName=collection_name,
            src_nodeID=src_node_id,
            dst_nodeIDs=dst_node_ids,
            sealed_segmentIDs=sealed_segment_ids,
        )
        return request

    @classmethod
    def manual_compaction(cls, collection_id, timetravel):
        if collection_id is None or not isinstance(collection_id, int):
            raise ParamError(message=f"collection_id value {collection_id} is illegal")

        if timetravel is None or not isinstance(timetravel, int):
            raise ParamError(message=f"timetravel value {timetravel} is illegal")

        request = milvus_types.ManualCompactionRequest()
        request.collectionID = collection_id
        request.timetravel = timetravel

        return request

    @classmethod
    def get_compaction_state(cls, compaction_id: int):
        if compaction_id is None or not isinstance(compaction_id, int):
            raise ParamError(message=f"compaction_id value {compaction_id} is illegal")

        request = milvus_types.GetCompactionStateRequest()
        request.compactionID = compaction_id
        return request

    @classmethod
    def get_compaction_state_with_plans(cls, compaction_id: int):
        if compaction_id is None or not isinstance(compaction_id, int):
            raise ParamError(message=f"compaction_id value {compaction_id} is illegal")

        request = milvus_types.GetCompactionPlansRequest()
        request.compactionID = compaction_id
        return request

    @classmethod
    def get_replicas(cls, collection_id: int):
        if collection_id is None or not isinstance(collection_id, int):
            raise ParamError(message=f"collection_id value {collection_id} is illegal")

        request = milvus_types.GetReplicasRequest(
            collectionID=collection_id,
            with_shard_nodes=True,
        )
        return request

    @classmethod
    def do_bulk_insert(cls, collection_name: str, partition_name: str, files: list, **kwargs):
        channel_names = kwargs.get("channel_names", None)
        req = milvus_types.ImportRequest(
            collection_name=collection_name,
            partition_name=partition_name,
            files=files,
        )
        if channel_names is not None:
            req.channel_names.extend(channel_names)

        for k, v in kwargs.items():
            if k in ("bucket",):
                kv_pair = common_types.KeyValuePair(key=str(k), value=str(v))
                req.options.append(kv_pair)

        return req

    @classmethod
    def get_bulk_insert_state(cls, task_id):
        if task_id is None or not isinstance(task_id, int):
            raise ParamError(f"task_id value {task_id} is not an integer")

        req = milvus_types.GetImportStateRequest(task=task_id)
        return req

    @classmethod
    def list_bulk_insert_tasks(cls, limit, collection_name):
        if limit is None or not isinstance(limit, int):
            raise ParamError(f"limit value {limit} is not an integer")

        request = milvus_types.ListImportTasksRequest(
            collection_name=collection_name,
            limit=limit,
        )
        return request

    @classmethod
    def create_user_request(cls, user, password):
        check_pass_param(user=user, password=password)
        return milvus_types.CreateCredentialRequest(username=user, password=base64.b64encode(password.encode('utf-8')))

    @classmethod
    def update_password_request(cls, user, old_password, new_password):
        check_pass_param(user=user)
        check_pass_param(password=old_password)
        check_pass_param(password=new_password)
        return milvus_types.UpdateCredentialRequest(username=user,
                                                    oldPassword=base64.b64encode(old_password.encode('utf-8')),
                                                    newPassword=base64.b64encode(new_password.encode('utf-8')),
                                                    )

    @classmethod
    def delete_user_request(cls, user):
        if not isinstance(user, str):
            raise ParamError(message=f"invalid user {user}")
        return milvus_types.DeleteCredentialRequest(username=user)

    @classmethod
    def list_usernames_request(cls):
        return milvus_types.ListCredUsersRequest()

    @classmethod
    def create_role_request(cls, role_name):
        check_pass_param(role_name=role_name)
        return milvus_types.CreateRoleRequest(entity=milvus_types.RoleEntity(name=role_name))

    @classmethod
    def drop_role_request(cls, role_name):
        check_pass_param(role_name=role_name)
        return milvus_types.DropRoleRequest(role_name=role_name)

    @classmethod
    def operate_user_role_request(cls, username, role_name, operate_user_role_type):
        check_pass_param(user=username)
        check_pass_param(role_name=role_name)
        check_pass_param(operate_user_role_type=operate_user_role_type)
        return milvus_types.OperateUserRoleRequest(username=username, role_name=role_name, type=operate_user_role_type)

    @classmethod
    def select_role_request(cls, role_name, include_user_info):
        if role_name:
            check_pass_param(role_name=role_name)
        check_pass_param(include_user_info=include_user_info)
        return milvus_types.SelectRoleRequest(role=milvus_types.RoleEntity(name=role_name) if role_name else None,
                                              include_user_info=include_user_info)

    @classmethod
    def select_user_request(cls, username, include_role_info):
        if username:
            check_pass_param(user=username)
        check_pass_param(include_role_info=include_role_info)
        return milvus_types.SelectUserRequest(user=milvus_types.UserEntity(name=username) if username else None,
                                              include_role_info=include_role_info)

    @classmethod
    def operate_privilege_request(cls, role_name, object, object_name, privilege, operate_privilege_type):
        check_pass_param(role_name=role_name)
        check_pass_param(object=object)
        check_pass_param(object_name=object_name)
        check_pass_param(privilege=privilege)
        check_pass_param(operate_privilege_type=operate_privilege_type)
        return milvus_types.OperatePrivilegeRequest(
            entity=milvus_types.GrantEntity(role=milvus_types.RoleEntity(name=role_name),
                                            object=milvus_types.ObjectEntity(name=object),
                                            object_name=object_name,
                                            grantor=milvus_types.GrantorEntity(
                                                privilege=milvus_types.PrivilegeEntity(name=privilege))),
            type=operate_privilege_type)

    @classmethod
    def select_grant_request(cls, role_name, object, object_name):
        check_pass_param(role_name=role_name)
        if object:
            check_pass_param(object=object)
        if object_name:
            check_pass_param(object_name=object_name)
        return milvus_types.SelectGrantRequest(
            entity=milvus_types.GrantEntity(role=milvus_types.RoleEntity(name=role_name),
                                            object=milvus_types.ObjectEntity(name=object) if object else None,
                                            object_name=object_name if object_name else None))

    @classmethod
    def get_server_version(cls):
        return milvus_types.GetVersionRequest()

    @classmethod
    def create_resource_group(cls, name):
        check_pass_param(resource_group_name=name)
        return milvus_types.CreateResourceGroupRequest(resource_group=name)

    @classmethod
    def drop_resource_group(cls, name):
        check_pass_param(resource_group_name=name)
        return milvus_types.DropResourceGroupRequest(resource_group=name)

    @classmethod
    def list_resource_groups(cls):
        return milvus_types.ListResourceGroupsRequest()

    @classmethod
    def describe_resource_group(cls, name):
        check_pass_param(resource_group_name=name)
        return milvus_types.DescribeResourceGroupRequest(resource_group=name)

    @classmethod
    def transfer_node(cls, source, target, num_node):
        check_pass_param(resource_group_name=source)
        check_pass_param(resource_group_name=target)
        return milvus_types.TransferNodeRequest(source_resource_group=source,
                                                target_resource_group=target,
                                                num_node=num_node)

    @classmethod
    def transfer_replica(cls, source, target, collection_name, num_replica):
        check_pass_param(resource_group_name=source)
        check_pass_param(resource_group_name=target)
        return milvus_types.TransferReplicaRequest(source_resource_group=source,
                                                   target_resource_group=target,
                                                   collection_name=collection_name,
                                                   num_replica=num_replica)

    @classmethod
    def flush_all_request(cls):
        return milvus_types.FlushAllRequest()

    @classmethod
    def get_flush_all_state_request(cls, flush_all_ts):
        return milvus_types.GetFlushAllStateRequest(flush_all_ts=flush_all_ts)



from . import common_pb2 as _common_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

BinaryVector: DataType
Bool: DataType
DESCRIPTOR: _descriptor.FileDescriptor
Double: DataType
FieldCreated: FieldState
FieldCreating: FieldState
FieldDropped: FieldState
FieldDropping: FieldState
Float: DataType
FloatVector: DataType
Int16: DataType
Int32: DataType
Int64: DataType
Int8: DataType
None: DataType
String: DataType
VarChar: DataType

class BoolArray(_message.Message):
    __slots__ = ["data"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[bool]
    def __init__(self, data: _Optional[_Iterable[bool]] = ...) -> None: ...

class BytesArray(_message.Message):
    __slots__ = ["data"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, data: _Optional[_Iterable[bytes]] = ...) -> None: ...

class CollectionSchema(_message.Message):
    __slots__ = ["autoID", "description", "fields", "name"]
    AUTOID_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    FIELDS_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    autoID: bool
    description: str
    fields: _containers.RepeatedCompositeFieldContainer[FieldSchema]
    name: str
    def __init__(self, name: _Optional[str] = ..., description: _Optional[str] = ..., autoID: bool = ..., fields: _Optional[_Iterable[_Union[FieldSchema, _Mapping]]] = ...) -> None: ...

class DoubleArray(_message.Message):
    __slots__ = ["data"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, data: _Optional[_Iterable[float]] = ...) -> None: ...

class FieldData(_message.Message):
    __slots__ = ["field_id", "field_name", "scalars", "type", "vectors"]
    FIELD_ID_FIELD_NUMBER: _ClassVar[int]
    FIELD_NAME_FIELD_NUMBER: _ClassVar[int]
    SCALARS_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    VECTORS_FIELD_NUMBER: _ClassVar[int]
    field_id: int
    field_name: str
    scalars: ScalarField
    type: DataType
    vectors: VectorField
    def __init__(self, type: _Optional[_Union[DataType, str]] = ..., field_name: _Optional[str] = ..., scalars: _Optional[_Union[ScalarField, _Mapping]] = ..., vectors: _Optional[_Union[VectorField, _Mapping]] = ..., field_id: _Optional[int] = ...) -> None: ...

class FieldSchema(_message.Message):
    __slots__ = ["autoID", "data_type", "description", "fieldID", "index_params", "is_primary_key", "name", "state", "type_params"]
    AUTOID_FIELD_NUMBER: _ClassVar[int]
    DATA_TYPE_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    FIELDID_FIELD_NUMBER: _ClassVar[int]
    INDEX_PARAMS_FIELD_NUMBER: _ClassVar[int]
    IS_PRIMARY_KEY_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    TYPE_PARAMS_FIELD_NUMBER: _ClassVar[int]
    autoID: bool
    data_type: DataType
    description: str
    fieldID: int
    index_params: _containers.RepeatedCompositeFieldContainer[_common_pb2.KeyValuePair]
    is_primary_key: bool
    name: str
    state: FieldState
    type_params: _containers.RepeatedCompositeFieldContainer[_common_pb2.KeyValuePair]
    def __init__(self, fieldID: _Optional[int] = ..., name: _Optional[str] = ..., is_primary_key: bool = ..., description: _Optional[str] = ..., data_type: _Optional[_Union[DataType, str]] = ..., type_params: _Optional[_Iterable[_Union[_common_pb2.KeyValuePair, _Mapping]]] = ..., index_params: _Optional[_Iterable[_Union[_common_pb2.KeyValuePair, _Mapping]]] = ..., autoID: bool = ..., state: _Optional[_Union[FieldState, str]] = ...) -> None: ...

class FloatArray(_message.Message):
    __slots__ = ["data"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, data: _Optional[_Iterable[float]] = ...) -> None: ...

class IDs(_message.Message):
    __slots__ = ["int_id", "str_id"]
    INT_ID_FIELD_NUMBER: _ClassVar[int]
    STR_ID_FIELD_NUMBER: _ClassVar[int]
    int_id: LongArray
    str_id: StringArray
    def __init__(self, int_id: _Optional[_Union[LongArray, _Mapping]] = ..., str_id: _Optional[_Union[StringArray, _Mapping]] = ...) -> None: ...

class IntArray(_message.Message):
    __slots__ = ["data"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, data: _Optional[_Iterable[int]] = ...) -> None: ...

class LongArray(_message.Message):
    __slots__ = ["data"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, data: _Optional[_Iterable[int]] = ...) -> None: ...

class ScalarField(_message.Message):
    __slots__ = ["bool_data", "bytes_data", "double_data", "float_data", "int_data", "long_data", "string_data"]
    BOOL_DATA_FIELD_NUMBER: _ClassVar[int]
    BYTES_DATA_FIELD_NUMBER: _ClassVar[int]
    DOUBLE_DATA_FIELD_NUMBER: _ClassVar[int]
    FLOAT_DATA_FIELD_NUMBER: _ClassVar[int]
    INT_DATA_FIELD_NUMBER: _ClassVar[int]
    LONG_DATA_FIELD_NUMBER: _ClassVar[int]
    STRING_DATA_FIELD_NUMBER: _ClassVar[int]
    bool_data: BoolArray
    bytes_data: BytesArray
    double_data: DoubleArray
    float_data: FloatArray
    int_data: IntArray
    long_data: LongArray
    string_data: StringArray
    def __init__(self, bool_data: _Optional[_Union[BoolArray, _Mapping]] = ..., int_data: _Optional[_Union[IntArray, _Mapping]] = ..., long_data: _Optional[_Union[LongArray, _Mapping]] = ..., float_data: _Optional[_Union[FloatArray, _Mapping]] = ..., double_data: _Optional[_Union[DoubleArray, _Mapping]] = ..., string_data: _Optional[_Union[StringArray, _Mapping]] = ..., bytes_data: _Optional[_Union[BytesArray, _Mapping]] = ...) -> None: ...

class SearchResultData(_message.Message):
    __slots__ = ["fields_data", "ids", "num_queries", "scores", "top_k", "topks"]
    FIELDS_DATA_FIELD_NUMBER: _ClassVar[int]
    IDS_FIELD_NUMBER: _ClassVar[int]
    NUM_QUERIES_FIELD_NUMBER: _ClassVar[int]
    SCORES_FIELD_NUMBER: _ClassVar[int]
    TOPKS_FIELD_NUMBER: _ClassVar[int]
    TOP_K_FIELD_NUMBER: _ClassVar[int]
    fields_data: _containers.RepeatedCompositeFieldContainer[FieldData]
    ids: IDs
    num_queries: int
    scores: _containers.RepeatedScalarFieldContainer[float]
    top_k: int
    topks: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, num_queries: _Optional[int] = ..., top_k: _Optional[int] = ..., fields_data: _Optional[_Iterable[_Union[FieldData, _Mapping]]] = ..., scores: _Optional[_Iterable[float]] = ..., ids: _Optional[_Union[IDs, _Mapping]] = ..., topks: _Optional[_Iterable[int]] = ...) -> None: ...

class StringArray(_message.Message):
    __slots__ = ["data"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, data: _Optional[_Iterable[str]] = ...) -> None: ...

class VectorField(_message.Message):
    __slots__ = ["binary_vector", "dim", "float_vector"]
    BINARY_VECTOR_FIELD_NUMBER: _ClassVar[int]
    DIM_FIELD_NUMBER: _ClassVar[int]
    FLOAT_VECTOR_FIELD_NUMBER: _ClassVar[int]
    binary_vector: bytes
    dim: int
    float_vector: FloatArray
    def __init__(self, dim: _Optional[int] = ..., float_vector: _Optional[_Union[FloatArray, _Mapping]] = ..., binary_vector: _Optional[bytes] = ...) -> None: ...

class DataType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class FieldState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
